
from typing import Tuple, Optional, Union
import pickle
class Tensor:
    def abs(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def abs_(self) -> 'Tensor': ...

    def acos(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def acos_(self) -> 'Tensor': ...

    @overload
    def add(self, other: 'Tensor', alpha: float=1, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def add(self, other: float, alpha: float=1, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def add_(self, other: 'Tensor', alpha: float=1) -> 'Tensor': ...

    @overload
    def add_(self, other: float, alpha: float=1) -> 'Tensor': ...

    def addbmm(self, batch1: 'Tensor', batch2: 'Tensor', beta: float=1, alpha: float=1, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def addbmm_(self, batch1: 'Tensor', batch2: 'Tensor', beta: float=1, alpha: float=1) -> 'Tensor': ...

    def addcdiv(self, tensor1: 'Tensor', tensor2: 'Tensor', value: float=1, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def addcdiv_(self, tensor1: 'Tensor', tensor2: 'Tensor', value: float=1) -> 'Tensor': ...

    def addcmul(self, tensor1: 'Tensor', tensor2: 'Tensor', value: float=1, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def addcmul_(self, tensor1: 'Tensor', tensor2: 'Tensor', value: float=1) -> 'Tensor': ...

    def addmm(self, mat1: 'Tensor', mat2: 'Tensor', beta: float=1, alpha: float=1, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def addmm_(self, mat1: 'Tensor', mat2: 'Tensor', beta: float=1, alpha: float=1) -> 'Tensor': ...

    def addmv(self, mat: 'Tensor', vec: 'Tensor', beta: float=1, alpha: float=1, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def addmv_(self, mat: 'Tensor', vec: 'Tensor', beta: float=1, alpha: float=1) -> 'Tensor': ...

    def addr(self, vec1: 'Tensor', vec2: 'Tensor', beta: float=1, alpha: float=1, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def addr_(self, vec1: 'Tensor', vec2: 'Tensor', beta: float=1, alpha: float=1) -> 'Tensor': ...

    @overload
    def all(self, dim: int, keepdim: bool=False, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def all(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def allclose(self, other: 'Tensor', rtol: float=1e-05, atol: float=1e-08, equal_nan: bool=False) -> bool: ...

    @overload
    def any(self, dim: int, keepdim: bool=False, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def any(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def argmax(self, dim=None, keepdim=False): ...

    def argmin(self, dim=None, keepdim=False): ...

    def argsort(self, dim=None, descending=False): ...

    def asin(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def asin_(self) -> 'Tensor': ...

    def atan(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def atan2(self, other: 'Tensor', *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def atan2_(self, other: 'Tensor') -> 'Tensor': ...

    def atan_(self) -> 'Tensor': ...

    def backward(self, gradient=None, retain_graph=None, create_graph=False): ...

    def baddbmm(self, batch1: 'Tensor', batch2: 'Tensor', beta: float=1, alpha: float=1, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def baddbmm_(self, batch1: 'Tensor', batch2: 'Tensor', beta: float=1, alpha: float=1) -> 'Tensor': ...

    @overload
    def bernoulli(self, generator: Generator=nullptr, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def bernoulli(self, p: float, generator: Generator=nullptr, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def bernoulli_(self, p: 'Tensor', generator: Generator=nullptr) -> 'Tensor': ...

    @overload
    def bernoulli_(self, p: float=0.5, generator: Generator=nullptr) -> 'Tensor': ...

    def bincount(self, weights: Optional['Tensor']={}, minlength: int=0) -> 'Tensor': ...

    def bmm(self, mat2: 'Tensor', *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def btrifact(self, info=None, pivot=True): ...

    def btrifact_with_info(self, pivot: bool=True, *, out: Optional['Tensor']=None) -> Tuple['Tensor', 'Tensor', 'Tensor']: ...

    def btrisolve(self, LU_data: 'Tensor', LU_pivots: 'Tensor', *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def cauchy_(self, median: float=0, sigma: float=1, generator: Generator=nullptr) -> 'Tensor': ...

    def ceil(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def ceil_(self) -> 'Tensor': ...

    def cholesky(self, upper: bool=False, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def chunk(self, chunks: int, dim: int=0) -> Tuple['Tensor', ...]: ...

    def clamp(self, min: Optional[float]=c10::nullopt, max: Optional[float]=c10::nullopt, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def clamp_(self, min: Optional[float]=c10::nullopt, max: Optional[float]=c10::nullopt) -> 'Tensor': ...

    def clone(self) -> 'Tensor': ...

    def contiguous(self) -> 'Tensor': ...

    def cos(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def cos_(self) -> 'Tensor': ...

    def cosh(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def cosh_(self) -> 'Tensor': ...

    def cross(self, other: 'Tensor', dim: int=-1, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def cumprod(self, dim: int, dtype: dtype, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def cumprod(self, dim: int, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def cumsum(self, dim: int, dtype: dtype, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def cumsum(self, dim: int, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def data_ptr(self) -> int: ...

    def dense_dim(self) -> int: ...

    def det(self) -> 'Tensor': ...

    def detach(self) -> 'Tensor': ...

    def detach_(self) -> 'Tensor': ...

    def diag(self, diagonal: int=0, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def diag_embed(self, offset: int=0, dim1: int=-2, dim2: int=-1) -> 'Tensor': ...

    def diagflat(self, offset: int=0) -> 'Tensor': ...

    def diagonal(self, offset: int=0, dim1: int=0, dim2: int=1) -> 'Tensor': ...

    def digamma(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def digamma_(self) -> 'Tensor': ...

    def dist(self, other: 'Tensor', p: float=2) -> 'Tensor': ...

    @overload
    def div(self, other: 'Tensor', *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def div(self, other: float, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def div_(self, other: 'Tensor') -> 'Tensor': ...

    @overload
    def div_(self, other: float) -> 'Tensor': ...

    def dot(self, tensor: 'Tensor', *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def eig(self, eigenvectors: bool=False, *, out: Optional['Tensor']=None) -> Tuple['Tensor', 'Tensor']: ...

    @overload
    def eq(self, other: float, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def eq(self, other: 'Tensor', *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def eq_(self, other: float) -> 'Tensor': ...

    @overload
    def eq_(self, other: 'Tensor') -> 'Tensor': ...

    def equal(self, other: 'Tensor') -> bool: ...

    def erf(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def erf_(self) -> 'Tensor': ...

    def erfc(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def erfc_(self) -> 'Tensor': ...

    def erfinv(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def erfinv_(self) -> 'Tensor': ...

    def exp(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def exp_(self) -> 'Tensor': ...

    def expand(self, size: Tuple[int, ...], implicit: bool=False) -> 'Tensor': ...

    def expand_as(self, other: 'Tensor') -> 'Tensor': ...

    def expm1(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def expm1_(self) -> 'Tensor': ...

    def exponential_(self, lambd: float=1, generator: Generator=nullptr) -> 'Tensor': ...

    def fft(self, signal_ndim: int, normalized: bool=False) -> 'Tensor': ...

    @overload
    def fill_(self, value: float) -> 'Tensor': ...

    @overload
    def fill_(self, value: 'Tensor') -> 'Tensor': ...

    def flatten(self, start_dim: int=0, end_dim: int=-1) -> 'Tensor': ...

    def flip(self, dims: Tuple[int, ...]) -> 'Tensor': ...

    def floor(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def floor_(self) -> 'Tensor': ...

    @overload
    def fmod(self, other: float, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def fmod(self, other: 'Tensor', *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def fmod_(self, other: float) -> 'Tensor': ...

    @overload
    def fmod_(self, other: 'Tensor') -> 'Tensor': ...

    def frac(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def frac_(self) -> 'Tensor': ...

    def gather(self, dim: int, index: 'Tensor', *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def ge(self, other: float, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def ge(self, other: 'Tensor', *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def ge_(self, other: float) -> 'Tensor': ...

    @overload
    def ge_(self, other: 'Tensor') -> 'Tensor': ...

    def gels(self, A: 'Tensor', *, out: Optional['Tensor']=None) -> Tuple['Tensor', 'Tensor']: ...

    def geometric_(self, p: float, generator: Generator=nullptr) -> 'Tensor': ...

    def geqrf(self, *, out: Optional['Tensor']=None) -> Tuple['Tensor', 'Tensor']: ...

    def ger(self, vec2: 'Tensor', *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def gesv(self, A: 'Tensor', *, out: Optional['Tensor']=None) -> Tuple['Tensor', 'Tensor']: ...

    @overload
    def gt(self, other: float, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def gt(self, other: 'Tensor', *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def gt_(self, other: float) -> 'Tensor': ...

    @overload
    def gt_(self, other: 'Tensor') -> 'Tensor': ...

    def hardshrink(self, lambd: float=0.5) -> 'Tensor': ...

    def histc(self, bins: int=100, min: float=0, max: float=0, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def ifft(self, signal_ndim: int, normalized: bool=False) -> 'Tensor': ...

    def index_add(self, dim, index, tensor): ...

    def index_add_(self, dim: int, index: 'Tensor', source: 'Tensor') -> 'Tensor': ...

    def index_copy(self, dim, index, tensor): ...

    def index_copy_(self, dim: int, index: 'Tensor', source: 'Tensor') -> 'Tensor': ...

    def index_fill(self, dim, index, value): ...

    @overload
    def index_fill_(self, dim: int, index: 'Tensor', value: float) -> 'Tensor': ...

    @overload
    def index_fill_(self, dim: int, index: 'Tensor', value: 'Tensor') -> 'Tensor': ...

    def index_put_(self, indices: Tuple['Tensor', ...], values: 'Tensor', accumulate: bool=False) -> 'Tensor': ...

    def index_select(self, dim: int, index: 'Tensor', *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def indices(self) -> 'Tensor': ...

    def inverse(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def irfft(self, signal_ndim: int, normalized: bool=False, onesided: bool=True, signal_sizes: Tuple[int, ...]={}) -> 'Tensor': ...

    def is_pinned(self): ...

    def is_set_to(self, tensor: 'Tensor') -> bool: ...

    def is_shared(self): ...

    def item(self) -> float: ...

    def kthvalue(self, k: int, dim: int=-1, keepdim: bool=False, *, out: Optional['Tensor']=None) -> Tuple['Tensor', 'Tensor']: ...

    @overload
    def le(self, other: float, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def le(self, other: 'Tensor', *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def le_(self, other: float) -> 'Tensor': ...

    @overload
    def le_(self, other: 'Tensor') -> 'Tensor': ...

    def lerp(self, end: 'Tensor', weight: float, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def lerp_(self, end: 'Tensor', weight: float) -> 'Tensor': ...

    def log(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def log10(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def log10_(self) -> 'Tensor': ...

    def log1p(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def log1p_(self) -> 'Tensor': ...

    def log2(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def log2_(self) -> 'Tensor': ...

    def log_(self) -> 'Tensor': ...

    def log_normal_(self, mean: float=1, std: float=2, generator: Generator=nullptr) -> 'Tensor': ...

    def logdet(self) -> 'Tensor': ...

    def logsumexp(self, dim: int, keepdim: bool=False, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def lt(self, other: float, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def lt(self, other: 'Tensor', *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def lt_(self, other: float) -> 'Tensor': ...

    @overload
    def lt_(self, other: 'Tensor') -> 'Tensor': ...

    def masked_fill(self, mask, value): ...

    @overload
    def masked_fill_(self, mask: 'Tensor', value: float) -> 'Tensor': ...

    @overload
    def masked_fill_(self, mask: 'Tensor', value: 'Tensor') -> 'Tensor': ...

    def masked_scatter(self, mask, tensor): ...

    def masked_scatter_(self, mask: 'Tensor', source: 'Tensor') -> 'Tensor': ...

    def masked_select(self, mask: 'Tensor', *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def matmul(self, other: 'Tensor', *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def matrix_power(self, n: int) -> 'Tensor': ...

    @overload
    def max(self, dim: int, keepdim: bool=False, *, out: Optional['Tensor']=None) -> Tuple['Tensor', 'Tensor']: ...

    @overload
    def max(self, other: 'Tensor', *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def max(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def mean(self, dtype: dtype, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def mean(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def mean(self, dim: Tuple[int, ...], keepdim: bool, dtype: dtype, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def mean(self, dim: Tuple[int, ...], keepdim: bool=False, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def mean(self, dim: Tuple[int, ...], dtype: dtype, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def median(self, dim: int, keepdim: bool=False, *, out: Optional['Tensor']=None) -> Tuple['Tensor', 'Tensor']: ...

    @overload
    def median(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def min(self, dim: int, keepdim: bool=False, *, out: Optional['Tensor']=None) -> Tuple['Tensor', 'Tensor']: ...

    @overload
    def min(self, other: 'Tensor', *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def min(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def mm(self, mat2: 'Tensor', *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def mode(self, dim: int=-1, keepdim: bool=False, *, out: Optional['Tensor']=None) -> Tuple['Tensor', 'Tensor']: ...

    @overload
    def mul(self, other: 'Tensor', *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def mul(self, other: float, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def mul_(self, other: 'Tensor') -> 'Tensor': ...

    @overload
    def mul_(self, other: float) -> 'Tensor': ...

    def multinomial(self, num_samples: int, replacement: bool=False, generator: Generator=nullptr, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def mv(self, vec: 'Tensor', *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def mvlgamma(self, p: int) -> 'Tensor': ...

    def mvlgamma_(self, p: int) -> 'Tensor': ...

    def narrow(self, dim: int, start: int, length: int) -> 'Tensor': ...

    def narrow_copy(self, dim: int, start: int, length: int) -> 'Tensor': ...

    @overload
    def ne(self, other: float, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def ne(self, other: 'Tensor', *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def ne_(self, other: float) -> 'Tensor': ...

    @overload
    def ne_(self, other: 'Tensor') -> 'Tensor': ...

    def neg(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def neg_(self) -> 'Tensor': ...

    def nonzero(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def norm(self, p='fro', dim=None, keepdim=False): ...

    def normal_(self, mean: float=0, std: float=1, generator: Generator=nullptr) -> 'Tensor': ...

    def numel(self) -> int: ...

    def orgqr(self, input2: 'Tensor', *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def ormqr(self, input2: 'Tensor', input3: 'Tensor', left: bool=True, transpose: bool=False, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def permute(self, dims: Tuple[int, ...]) -> 'Tensor': ...

    def pinverse(self, rcond: float=1e-15) -> 'Tensor': ...

    def potrf(self, upper=True): ...

    def potri(self, upper: bool=True, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def potrs(self, input2: 'Tensor', upper: bool=True, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def pow(self, exponent: float, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def pow(self, exponent: 'Tensor', *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def pow(self, exponent: 'Tensor', *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def pow_(self, exponent: float) -> 'Tensor': ...

    @overload
    def pow_(self, exponent: 'Tensor') -> 'Tensor': ...

    @overload
    def prod(self, dtype: dtype, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def prod(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def prod(self, dim: int, keepdim: bool, dtype: dtype, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def prod(self, dim: int, keepdim: bool=False, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def prod(self, dim: int, dtype: dtype, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def pstrf(self, upper: bool=True, tol: float=-1, *, out: Optional['Tensor']=None) -> Tuple['Tensor', 'Tensor']: ...

    def put_(self, index: 'Tensor', source: 'Tensor', accumulate: bool=False) -> 'Tensor': ...

    def qr(self, *, out: Optional['Tensor']=None) -> Tuple['Tensor', 'Tensor']: ...

    @overload
    def random_(self, from_: int, to: int, generator: Generator=nullptr) -> 'Tensor': ...

    @overload
    def random_(self, to: int, generator: Generator=nullptr) -> 'Tensor': ...

    @overload
    def random_(self, generator: Generator=nullptr) -> 'Tensor': ...

    def reciprocal(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def reciprocal_(self) -> 'Tensor': ...

    def register_hook(self, hook): ...

    @overload
    def remainder(self, other: float, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def remainder(self, other: 'Tensor', *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def remainder_(self, other: float) -> 'Tensor': ...

    @overload
    def remainder_(self, other: 'Tensor') -> 'Tensor': ...

    def renorm(self, p: float, dim: int, maxnorm: float, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def renorm_(self, p: float, dim: int, maxnorm: float) -> 'Tensor': ...

    def repeat(self, repeats: Tuple[int, ...]) -> 'Tensor': ...

    def reshape(self, shape: Tuple[int, ...]) -> 'Tensor': ...

    def reshape_as(self, other: 'Tensor') -> 'Tensor': ...

    def resize_(self, size: Tuple[int, ...]) -> 'Tensor': ...

    def resize_as_(self, the_template: 'Tensor') -> 'Tensor': ...

    def retain_grad(self): ...

    def rfft(self, signal_ndim: int, normalized: bool=False, onesided: bool=True) -> 'Tensor': ...

    def roll(self, shifts: Tuple[int, ...], dims: Tuple[int, ...]={}) -> 'Tensor': ...

    def rot90(self, k: int=1, dims: Tuple[int, ...]={0,1}) -> 'Tensor': ...

    def round(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def round_(self) -> 'Tensor': ...

    def rsqrt(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def rsqrt_(self) -> 'Tensor': ...

    def scatter(self, dim, index, source): ...

    @overload
    def scatter_(self, dim: int, index: 'Tensor', src: 'Tensor') -> 'Tensor': ...

    @overload
    def scatter_(self, dim: int, index: 'Tensor', value: float) -> 'Tensor': ...

    def scatter_add(self, dim, index, source): ...

    def scatter_add_(self, dim: int, index: 'Tensor', src: 'Tensor') -> 'Tensor': ...

    def select(self, dim: int, index: int) -> 'Tensor': ...

    @overload
    def set_(self, source: Storage) -> 'Tensor': ...

    @overload
    def set_(self, source: Storage, storage_offset: int, size: Tuple[int, ...], stride: Tuple[int, ...]={}) -> 'Tensor': ...

    @overload
    def set_(self, source: 'Tensor') -> 'Tensor': ...

    @overload
    def set_(self) -> 'Tensor': ...

    def share_memory_(self): ...

    def sigmoid(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def sigmoid_(self) -> 'Tensor': ...

    def sign(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def sign_(self) -> 'Tensor': ...

    def sin(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def sin_(self) -> 'Tensor': ...

    def sinh(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def sinh_(self) -> 'Tensor': ...

    def size(self, dim: int) -> int: ...

    def slogdet(self) -> Tuple['Tensor', 'Tensor']: ...

    def sort(self, dim: int=-1, descending: bool=False, *, out: Optional['Tensor']=None) -> Tuple['Tensor', 'Tensor']: ...

    def sparse_dim(self) -> int: ...

    def sparse_mask(self, mask: 'Tensor') -> 'Tensor': ...

    def split(self, split_size, dim=0): ...

    def sqrt(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def sqrt_(self) -> 'Tensor': ...

    @overload
    def squeeze(self) -> 'Tensor': ...

    @overload
    def squeeze(self, dim: int) -> 'Tensor': ...

    @overload
    def squeeze_(self) -> 'Tensor': ...

    @overload
    def squeeze_(self, dim: int) -> 'Tensor': ...

    @overload
    def std(self, unbiased: bool=True, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def std(self, dim: int, unbiased: bool=True, keepdim: bool=False, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def stft(self, n_fft, hop_length=None, win_length=None, window=None, center=True, pad_mode='reflect', normalized=False, onesided=True): ...

    def stride(self, dim: int) -> int: ...

    @overload
    def sub(self, other: 'Tensor', alpha: float=1, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def sub(self, other: float, alpha: float=1, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def sub_(self, other: 'Tensor', alpha: float=1) -> 'Tensor': ...

    @overload
    def sub_(self, other: float, alpha: float=1) -> 'Tensor': ...

    @overload
    def sum(self, dtype: dtype, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def sum(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def sum(self, dim: Tuple[int, ...], keepdim: bool, dtype: dtype, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def sum(self, dim: Tuple[int, ...], keepdim: bool=False, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def sum(self, dim: Tuple[int, ...], dtype: dtype, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def svd(self, some: bool=True, compute_uv: bool=True, *, out: Optional['Tensor']=None) -> Tuple['Tensor', 'Tensor', 'Tensor']: ...

    def symeig(self, eigenvectors: bool=False, upper: bool=True, *, out: Optional['Tensor']=None) -> Tuple['Tensor', 'Tensor']: ...

    def t(self) -> 'Tensor': ...

    def t_(self) -> 'Tensor': ...

    def take(self, index: 'Tensor', *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def tan_(self) -> 'Tensor': ...

    def tanh(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def tanh_(self) -> 'Tensor': ...

    @overload
    def to(self, non_blocking: bool=False, copy: bool=False, *, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> 'Tensor': ...

    @overload
    def to(self, device: Union[device, str, None], dtype: dtype, non_blocking: bool=False, copy: bool=False) -> 'Tensor': ...

    @overload
    def to(self, dtype: dtype, non_blocking: bool=False, copy: bool=False) -> 'Tensor': ...

    @overload
    def to(self, other: 'Tensor', non_blocking: bool=False, copy: bool=False) -> 'Tensor': ...

    @overload
    def to_sparse(self, sparse_dim: int) -> 'Tensor': ...

    @overload
    def to_sparse(self) -> 'Tensor': ...

    def topk(self, k: int, dim: int=-1, largest: bool=True, sorted: bool=True, *, out: Optional['Tensor']=None) -> Tuple['Tensor', 'Tensor']: ...

    def trace(self) -> 'Tensor': ...

    def transpose(self, dim0: int, dim1: int) -> 'Tensor': ...

    def transpose_(self, dim0: int, dim1: int) -> 'Tensor': ...

    def tril(self, diagonal: int=0, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def tril_(self, diagonal: int=0) -> 'Tensor': ...

    def triu(self, diagonal: int=0, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def triu_(self, diagonal: int=0) -> 'Tensor': ...

    def trtrs(self, A: 'Tensor', upper: bool=True, transpose: bool=False, unitriangular: bool=False, *, out: Optional['Tensor']=None) -> Tuple['Tensor', 'Tensor']: ...

    def trunc(self, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def trunc_(self) -> 'Tensor': ...

    def type_as(self, other: 'Tensor') -> 'Tensor': ...

    def unbind(self, dim: int=0) -> Tuple['Tensor', ...]: ...

    def unfold(self, dimension: int, size: int, step: int) -> 'Tensor': ...

    def uniform_(self, from_: float=0, to: float=1, generator: Generator=nullptr) -> 'Tensor': ...

    def unique(self, sorted=False, return_inverse=False, dim=None): ...

    def unsqueeze(self, dim: int) -> 'Tensor': ...

    def unsqueeze_(self, dim: int) -> 'Tensor': ...

    def values(self) -> 'Tensor': ...

    @overload
    def var(self, unbiased: bool=True, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    @overload
    def var(self, dim: int, unbiased: bool=True, keepdim: bool=False, *, out: Optional['Tensor']=None) -> 'Tensor': ...

    def view(self, size: Tuple[int, ...]) -> 'Tensor': ...

    def view_as(self, other: 'Tensor') -> 'Tensor': ...

    def where(self, condition: 'Tensor', other: 'Tensor') -> 'Tensor': ...

    def zero_(self) -> 'Tensor': ...


def abs(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def acos(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def adaptive_avg_pool1d(self: Tensor, output_size: Tuple[int, ...]) -> Tensor: ...

@overload
def add(self: Tensor, other: Tensor, alpha: float=1, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def add(self: Tensor, other: float, alpha: float=1, *, out: Optional[Tensor]=None) -> Tensor: ...

def addbmm(self: Tensor, batch1: Tensor, batch2: Tensor, beta: float=1, alpha: float=1, *, out: Optional[Tensor]=None) -> Tensor: ...

def addcdiv(self: Tensor, tensor1: Tensor, tensor2: Tensor, value: float=1, *, out: Optional[Tensor]=None) -> Tensor: ...

def addcmul(self: Tensor, tensor1: Tensor, tensor2: Tensor, value: float=1, *, out: Optional[Tensor]=None) -> Tensor: ...

def addmm(self: Tensor, mat1: Tensor, mat2: Tensor, beta: float=1, alpha: float=1, *, out: Optional[Tensor]=None) -> Tensor: ...

def addmv(self: Tensor, mat: Tensor, vec: Tensor, beta: float=1, alpha: float=1, *, out: Optional[Tensor]=None) -> Tensor: ...

def addr(self: Tensor, vec1: Tensor, vec2: Tensor, beta: float=1, alpha: float=1, *, out: Optional[Tensor]=None) -> Tensor: ...

def allclose(self: Tensor, other: Tensor, rtol: float=1e-05, atol: float=1e-08, equal_nan: bool=False) -> bool: ...

@overload
def arange(start: float, end: float, *, out: Optional[Tensor]=None, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def arange(start: float, end: float, step: float, *, out: Optional[Tensor]=None, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def arange(end: float, *, out: Optional[Tensor]=None, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

def argmax(input, dim=None, keepdim=False): ...

def argmin(input, dim=None, keepdim=False): ...

def argsort(input, dim=None, descending=False): ...

def asin(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def atan(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def atan2(self: Tensor, other: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def avg_pool1d(self: Tensor, kernel_size: Tuple[int, ...], stride: Tuple[int, ...]={}, padding: Tuple[int, ...]=0, ceil_mode: bool=False, count_include_pad: bool=True) -> Tensor: ...

def baddbmm(self: Tensor, batch1: Tensor, batch2: Tensor, beta: float=1, alpha: float=1, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def bartlett_window(window_length: int, *, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def bartlett_window(window_length: int, periodic: bool, *, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def bernoulli(self: Tensor, generator: Generator=nullptr, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def bernoulli(self: Tensor, p: float, generator: Generator=nullptr, *, out: Optional[Tensor]=None) -> Tensor: ...

def bincount(self: Tensor, weights: Optional[Tensor]={}, minlength: int=0) -> Tensor: ...

@overload
def blackman_window(window_length: int, *, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def blackman_window(window_length: int, periodic: bool, *, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

def bmm(self: Tensor, mat2: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def broadcast_tensors(*tensors:Tensor) -> List[Tensor]: ...

def btrifact(A:Tensor, info:Union[Tensor, NoneType]=None, pivot:bool=True) -> Tuple[Tensor, Tensor]: ...

def btrifact_with_info(self: Tensor, pivot: bool=True, *, out: Optional[Tensor]=None) -> Tuple[Tensor, Tensor, Tensor]: ...

def btrisolve(self: Tensor, LU_data: Tensor, LU_pivots: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def btriunpack(LU_data:Tensor, LU_pivots:Tensor, unpack_data:bool=True, unpack_pivots:bool=True) -> Tuple[Union[Tensor, NoneType], Union[Tensor, NoneType], Union[Tensor, NoneType]]: ...

def cat(tensors: Tuple[Tensor, ...], dim: int=0, *, out: Optional[Tensor]=None) -> Tensor: ...

def ceil(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def celu_(self: Tensor, alpha: float=1.0) -> Tensor: ...

def chain_matmul(*matrices): ...

def cholesky(self: Tensor, upper: bool=False, *, out: Optional[Tensor]=None) -> Tensor: ...

def chunk(self: Tensor, chunks: int, dim: int=0) -> Tuple[Tensor, ...]: ...

def clamp(self: Tensor, min: Optional[float]=c10::nullopt, max: Optional[float]=c10::nullopt, *, out: Optional[Tensor]=None) -> Tensor: ...

def compiled_with_cxx11_abi(): ...

def conv1d(input: Tensor, weight: Tensor, bias: Optional[Tensor]={}, stride: Tuple[int, ...]=1, padding: Tuple[int, ...]=0, dilation: Tuple[int, ...]=1, groups: int=1) -> Tensor: ...

def conv2d(input: Tensor, weight: Tensor, bias: Optional[Tensor]={}, stride: Tuple[int, ...]=1, padding: Tuple[int, ...]=0, dilation: Tuple[int, ...]=1, groups: int=1) -> Tensor: ...

def conv3d(input: Tensor, weight: Tensor, bias: Optional[Tensor]={}, stride: Tuple[int, ...]=1, padding: Tuple[int, ...]=0, dilation: Tuple[int, ...]=1, groups: int=1) -> Tensor: ...

def conv_tbc(self: Tensor, weight: Tensor, bias: Tensor, pad: int=0) -> Tensor: ...

def conv_transpose1d(input: Tensor, weight: Tensor, bias: Optional[Tensor]={}, stride: Tuple[int, ...]=1, padding: Tuple[int, ...]=0, output_padding: Tuple[int, ...]=0, groups: int=1, dilation: Tuple[int, ...]=1) -> Tensor: ...

def conv_transpose2d(input: Tensor, weight: Tensor, bias: Optional[Tensor]={}, stride: Tuple[int, ...]=1, padding: Tuple[int, ...]=0, output_padding: Tuple[int, ...]=0, groups: int=1, dilation: Tuple[int, ...]=1) -> Tensor: ...

def conv_transpose3d(input: Tensor, weight: Tensor, bias: Optional[Tensor]={}, stride: Tuple[int, ...]=1, padding: Tuple[int, ...]=0, output_padding: Tuple[int, ...]=0, groups: int=1, dilation: Tuple[int, ...]=1) -> Tensor: ...

def cos(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def cosh(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def cosine_similarity(x1: Tensor, x2: Tensor, dim: int=1, eps: float=1e-08) -> Tensor: ...

def cross(self: Tensor, other: Tensor, dim: int=-1, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def cumprod(self: Tensor, dim: int, dtype: dtype, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def cumprod(self: Tensor, dim: int, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def cumsum(self: Tensor, dim: int, dtype: dtype, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def cumsum(self: Tensor, dim: int, *, out: Optional[Tensor]=None) -> Tensor: ...

def det(self: Tensor) -> Tensor: ...

def diag(self: Tensor, diagonal: int=0, *, out: Optional[Tensor]=None) -> Tensor: ...

def diag_embed(self: Tensor, offset: int=0, dim1: int=-2, dim2: int=-1) -> Tensor: ...

def diagflat(self: Tensor, offset: int=0) -> Tensor: ...

def diagonal(self: Tensor, offset: int=0, dim1: int=0, dim2: int=1) -> Tensor: ...

def digamma(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def dist(self: Tensor, other: Tensor, p: float=2) -> Tensor: ...

@overload
def div(self: Tensor, other: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def div(self: Tensor, other: float, *, out: Optional[Tensor]=None) -> Tensor: ...

def dot(self: Tensor, tensor: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def eig(self: Tensor, eigenvectors: bool=False, *, out: Optional[Tensor]=None) -> Tuple[Tensor, Tensor]: ...

def einsum(equation:str, *operands:Tensor) -> Tensor: ...

@overload
def empty(size: Tuple[int, ...], *, out: Optional[Tensor]=None, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def empty(*size: int, out: Optional[Tensor]=None, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def empty_like(self: Tensor) -> Tensor: ...

@overload
def empty_like(self: Tensor, *, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def eq(self: Tensor, other: float, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def eq(self: Tensor, other: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def equal(self: Tensor, other: Tensor) -> bool: ...

def erf(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def erfc(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def erfinv(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def exp(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def expm1(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def eye(n: int, *, out: Optional[Tensor]=None, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def eye(n: int, m: int, *, out: Optional[Tensor]=None, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

def fft(self: Tensor, signal_ndim: int, normalized: bool=False) -> Tensor: ...

def flatten(self: Tensor, start_dim: int=0, end_dim: int=-1) -> Tensor: ...

def flip(self: Tensor, dims: Tuple[int, ...]) -> Tensor: ...

def floor(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def fmod(self: Tensor, other: float, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def fmod(self: Tensor, other: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def frac(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def full(size: Tuple[int, ...], fill_value: float, *, out: Optional[Tensor]=None, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def full_like(self: Tensor, fill_value: float) -> Tensor: ...

@overload
def full_like(self: Tensor, fill_value: float, *, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

def gather(self: Tensor, dim: int, index: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def ge(self: Tensor, other: float, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def ge(self: Tensor, other: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def gels(self: Tensor, A: Tensor, *, out: Optional[Tensor]=None) -> Tuple[Tensor, Tensor]: ...

def geqrf(self: Tensor, *, out: Optional[Tensor]=None) -> Tuple[Tensor, Tensor]: ...

def ger(self: Tensor, vec2: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def gesv(self: Tensor, A: Tensor, *, out: Optional[Tensor]=None) -> Tuple[Tensor, Tensor]: ...

def get_rng_state(): ...

@overload
def gt(self: Tensor, other: float, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def gt(self: Tensor, other: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def hamming_window(window_length: int, *, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def hamming_window(window_length: int, periodic: bool, *, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def hamming_window(window_length: int, periodic: bool, alpha: float, *, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def hamming_window(window_length: int, periodic: bool, alpha: float, beta: float, *, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def hann_window(window_length: int, *, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def hann_window(window_length: int, periodic: bool, *, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

def histc(self: Tensor, bins: int=100, min: float=0, max: float=0, *, out: Optional[Tensor]=None) -> Tensor: ...

def ifft(self: Tensor, signal_ndim: int, normalized: bool=False) -> Tensor: ...

def index_select(self: Tensor, dim: int, index: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def initial_seed(): ...

def inverse(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def irfft(self: Tensor, signal_ndim: int, normalized: bool=False, onesided: bool=True, signal_sizes: Tuple[int, ...]={}) -> Tensor: ...

def is_storage(obj): ...

def is_tensor(obj): ...

def isfinite(tensor): ...

def isinf(tensor): ...

def isnan(tensor): ...

def kthvalue(self: Tensor, k: int, dim: int=-1, keepdim: bool=False, *, out: Optional[Tensor]=None) -> Tuple[Tensor, Tensor]: ...

@overload
def le(self: Tensor, other: float, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def le(self: Tensor, other: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def lerp(self: Tensor, end: Tensor, weight: float, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def linspace(start: float, end: float, *, out: Optional[Tensor]=None, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def linspace(start: float, end: float, steps: int, *, out: Optional[Tensor]=None, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

def load(f, map_location=None, pickle_module=pickle): ...

def log(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def log10(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def log1p(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def log2(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def logdet(self: Tensor) -> Tensor: ...

@overload
def logspace(start: float, end: float, *, out: Optional[Tensor]=None, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def logspace(start: float, end: float, steps: int, *, out: Optional[Tensor]=None, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

def logsumexp(self: Tensor, dim: int, keepdim: bool=False, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def lt(self: Tensor, other: float, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def lt(self: Tensor, other: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def manual_seed(seed): ...

def masked_select(self: Tensor, mask: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def matmul(self: Tensor, other: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def matrix_power(self: Tensor, n: int) -> Tensor: ...

@overload
def matrix_rank(self: Tensor, tol: float, symmetric: bool=False) -> Tensor: ...

@overload
def matrix_rank(self: Tensor, symmetric: bool=False) -> Tensor: ...

@overload
def max(self: Tensor, dim: int, keepdim: bool=False, *, out: Optional[Tensor]=None) -> Tuple[Tensor, Tensor]: ...

@overload
def max(self: Tensor, other: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def max(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def mean(self: Tensor, dtype: dtype, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def mean(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def mean(self: Tensor, dim: Tuple[int, ...], keepdim: bool, dtype: dtype, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def mean(self: Tensor, dim: Tuple[int, ...], keepdim: bool=False, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def mean(self: Tensor, dim: Tuple[int, ...], dtype: dtype, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def median(self: Tensor, dim: int, keepdim: bool=False, *, out: Optional[Tensor]=None) -> Tuple[Tensor, Tensor]: ...

@overload
def median(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def meshgrid(*tensors, **kwargs): ...

@overload
def min(self: Tensor, dim: int, keepdim: bool=False, *, out: Optional[Tensor]=None) -> Tuple[Tensor, Tensor]: ...

@overload
def min(self: Tensor, other: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def min(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def mm(self: Tensor, mat2: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def mode(self: Tensor, dim: int=-1, keepdim: bool=False, *, out: Optional[Tensor]=None) -> Tuple[Tensor, Tensor]: ...

@overload
def mul(self: Tensor, other: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def mul(self: Tensor, other: float, *, out: Optional[Tensor]=None) -> Tensor: ...

def multinomial(self: Tensor, num_samples: int, replacement: bool=False, generator: Generator=nullptr, *, out: Optional[Tensor]=None) -> Tensor: ...

def mv(self: Tensor, vec: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def mvlgamma(self: Tensor, p: int) -> Tensor: ...

def narrow(self: Tensor, dim: int, start: int, length: int) -> Tensor: ...

@overload
def ne(self: Tensor, other: float, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def ne(self: Tensor, other: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def neg(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def nonzero(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def norm(input, p='fro', dim=None, keepdim=False, out=None): ...

@overload
def normal(mean: Tensor, std: float=1, generator: Generator=nullptr, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def normal(mean: float, std: Tensor, generator: Generator=nullptr, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def normal(mean: Tensor, std: Tensor, generator: Generator=nullptr, *, out: Optional[Tensor]=None) -> Tensor: ...

def numel(self: Tensor) -> int: ...

@overload
def ones(size: Tuple[int, ...], *, out: Optional[Tensor]=None, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def ones(*size: int, out: Optional[Tensor]=None, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def ones_like(self: Tensor) -> Tensor: ...

@overload
def ones_like(self: Tensor, *, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

def orgqr(self: Tensor, input2: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def ormqr(self: Tensor, input2: Tensor, input3: Tensor, left: bool=True, transpose: bool=False, *, out: Optional[Tensor]=None) -> Tensor: ...

def pdist(self: Tensor, p: float=2) -> Tensor: ...

def pinverse(self: Tensor, rcond: float=1e-15) -> Tensor: ...

def pixel_shuffle(self: Tensor, upscale_factor: int) -> Tensor: ...

def potrf(a, upper=True, out=None): ...

def potri(self: Tensor, upper: bool=True, *, out: Optional[Tensor]=None) -> Tensor: ...

def potrs(self: Tensor, input2: Tensor, upper: bool=True, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def pow(self: Tensor, exponent: float, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def pow(self: Tensor, exponent: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def pow(self: float, exponent: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def prod(self: Tensor, dtype: dtype, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def prod(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def prod(self: Tensor, dim: int, keepdim: bool, dtype: dtype, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def prod(self: Tensor, dim: int, keepdim: bool=False, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def prod(self: Tensor, dim: int, dtype: dtype, *, out: Optional[Tensor]=None) -> Tensor: ...

def pstrf(self: Tensor, upper: bool=True, tol: float=-1, *, out: Optional[Tensor]=None) -> Tuple[Tensor, Tensor]: ...

def qr(self: Tensor, *, out: Optional[Tensor]=None) -> Tuple[Tensor, Tensor]: ...

@overload
def rand(size: Tuple[int, ...], *, out: Optional[Tensor]=None, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def rand(*size: int, out: Optional[Tensor]=None, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def rand(size: Tuple[int, ...], generator: Generator, *, out: Optional[Tensor]=None, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def rand_like(self: Tensor) -> Tensor: ...

@overload
def rand_like(self: Tensor, *, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def randint(high: int, size: Tuple[int, ...], *, out: Optional[Tensor]=None, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def randint(high: int, size: Tuple[int, ...], generator: Generator, *, out: Optional[Tensor]=None, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def randint(low: int, high: int, size: Tuple[int, ...], *, out: Optional[Tensor]=None, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def randint(low: int, high: int, size: Tuple[int, ...], generator: Generator, *, out: Optional[Tensor]=None, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def randint_like(self: Tensor, high: int) -> Tensor: ...

@overload
def randint_like(self: Tensor, low: int, high: int) -> Tensor: ...

@overload
def randint_like(self: Tensor, high: int, *, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def randint_like(self: Tensor, low: int, high: int, *, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def randn(size: Tuple[int, ...], *, out: Optional[Tensor]=None, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def randn(*size: int, out: Optional[Tensor]=None, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def randn(size: Tuple[int, ...], generator: Generator, *, out: Optional[Tensor]=None, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def randn_like(self: Tensor) -> Tensor: ...

@overload
def randn_like(self: Tensor, *, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def randperm(n: int, *, out: Optional[Tensor]=None, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def randperm(n: int, generator: Generator, *, out: Optional[Tensor]=None, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def range(start: float, end: float, *, out: Optional[Tensor]=None, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def range(start: float, end: float, step: float, *, out: Optional[Tensor]=None, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

def reciprocal(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def relu_(self: Tensor) -> Tensor: ...

@overload
def remainder(self: Tensor, other: float, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def remainder(self: Tensor, other: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def renorm(self: Tensor, p: float, dim: int, maxnorm: float, *, out: Optional[Tensor]=None) -> Tensor: ...

def reshape(self: Tensor, shape: Tuple[int, ...]) -> Tensor: ...

def rfft(self: Tensor, signal_ndim: int, normalized: bool=False, onesided: bool=True) -> Tensor: ...

def roll(self: Tensor, shifts: Tuple[int, ...], dims: Tuple[int, ...]={}) -> Tensor: ...

def rot90(self: Tensor, k: int=1, dims: Tuple[int, ...]={0,1}) -> Tensor: ...

def round(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def rrelu_(self: Tensor, lower: float=0.125, upper: float=0.3333333333333333, training: bool=False, generator: Generator=nullptr) -> Tensor: ...

def rsqrt(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def save(obj, f, pickle_module=pickle, pickle_protocol=2): ...

def selu_(self: Tensor) -> Tensor: ...

def set_default_dtype(d): ...

def set_default_tensor_type(t): ...

def set_printoptions(precision=None, threshold=None, edgeitems=None, linewidth=None, profile=None): ...

def set_rng_state(new_state): ...

def sigmoid(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def sign(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def sin(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def sinh(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def slogdet(self: Tensor) -> Tuple[Tensor, Tensor]: ...

def sort(self: Tensor, dim: int=-1, descending: bool=False, *, out: Optional[Tensor]=None) -> Tuple[Tensor, Tensor]: ...

@overload
def sparse_coo_tensor(size: Tuple[int, ...], *, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def sparse_coo_tensor(*size: int, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def sparse_coo_tensor(indices: Tensor, values: Tensor, *, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def sparse_coo_tensor(indices: Tensor, values: Tensor, size: Tuple[int, ...], *, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

def split(tensor:Tensor, split_size_or_sections:Union[List[int], int], dim:int=0) -> List[Tensor]: ...

def sqrt(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def squeeze(self: Tensor) -> Tensor: ...

@overload
def squeeze(self: Tensor, dim: int) -> Tensor: ...

def stack(tensors: Tuple[Tensor, ...], dim: int=0, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def std(self: Tensor, unbiased: bool=True, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def std(self: Tensor, dim: int, unbiased: bool=True, keepdim: bool=False, *, out: Optional[Tensor]=None) -> Tensor: ...

def stft(input, n_fft, hop_length=None, win_length=None, window=None, center=True, pad_mode='reflect', normalized=False, onesided=True): ...

@overload
def sum(self: Tensor, dtype: dtype, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def sum(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def sum(self: Tensor, dim: Tuple[int, ...], keepdim: bool, dtype: dtype, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def sum(self: Tensor, dim: Tuple[int, ...], keepdim: bool=False, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def sum(self: Tensor, dim: Tuple[int, ...], dtype: dtype, *, out: Optional[Tensor]=None) -> Tensor: ...

def svd(self: Tensor, some: bool=True, compute_uv: bool=True, *, out: Optional[Tensor]=None) -> Tuple[Tensor, Tensor, Tensor]: ...

def symeig(self: Tensor, eigenvectors: bool=False, upper: bool=True, *, out: Optional[Tensor]=None) -> Tuple[Tensor, Tensor]: ...

def t(self: Tensor) -> Tensor: ...

def take(self: Tensor, index: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def tan(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def tanh(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def tensordot(a, b, dims=2): ...

def threshold_(self: Tensor, threshold: float, value: float) -> Tensor: ...

def topk(self: Tensor, k: int, dim: int=-1, largest: bool=True, sorted: bool=True, *, out: Optional[Tensor]=None) -> Tuple[Tensor, Tensor]: ...

def trace(self: Tensor) -> Tensor: ...

def transpose(self: Tensor, dim0: int, dim1: int) -> Tensor: ...

def tril(self: Tensor, diagonal: int=0, *, out: Optional[Tensor]=None) -> Tensor: ...

def triu(self: Tensor, diagonal: int=0, *, out: Optional[Tensor]=None) -> Tensor: ...

def trtrs(self: Tensor, A: Tensor, upper: bool=True, transpose: bool=False, unitriangular: bool=False, *, out: Optional[Tensor]=None) -> Tuple[Tensor, Tensor]: ...

def trunc(self: Tensor, *, out: Optional[Tensor]=None) -> Tensor: ...

def unbind(self: Tensor, dim: int=0) -> Tuple[Tensor, ...]: ...

def unique(input, sorted=False, return_inverse=False, dim=None): ...

def unsqueeze(self: Tensor, dim: int) -> Tensor: ...

@overload
def var(self: Tensor, unbiased: bool=True, *, out: Optional[Tensor]=None) -> Tensor: ...

@overload
def var(self: Tensor, dim: int, unbiased: bool=True, keepdim: bool=False, *, out: Optional[Tensor]=None) -> Tensor: ...

def where(condition: Tensor, self: Tensor, other: Tensor) -> Tensor: ...

@overload
def zeros(size: Tuple[int, ...], *, out: Optional[Tensor]=None, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def zeros(*size: int, out: Optional[Tensor]=None, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...

@overload
def zeros_like(self: Tensor) -> Tensor: ...

@overload
def zeros_like(self: Tensor, *, dtype: dtype=None, layout: layout=torch.strided, device: device=None, requires_grad:bool=False) -> Tensor: ...




from .parallel_apply import parallel_apply
from .replicate import replicate
from .data_parallel import DataParallel, data_parallel
from .scatter_gather import scatter, gather
from .distributed import DistributedDataParallel
from .distributed_cpu import DistributedDataParallelCPU
import torch.nn.parallel.deprecated

__all__ = ['replicate', 'scatter', 'parallel_apply', 'gather', 'data_parallel',
           'DataParallel', 'DistributedDataParallel', 'DistributedDataParallelCPU']



# @package layer_model_helper
# Module caffe2.python.layer_model_helper
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from caffe2.python import core, model_helper, schema, scope, utils, muji
from caffe2.python.modeling.parameter_info import (
    ParameterInfo,
)
from caffe2.python.modeling.parameter_sharing import (
    parameter_sharing_context,
)
from caffe2.python.modeling.net_modifier import NetModifier

from caffe2.python.optimizer import get_param_device
from caffe2.python.regularizer import Regularizer, RegularizationBy
from caffe2.python.layers import layers
from caffe2.proto import caffe2_pb2
from future.utils import viewitems, viewvalues

import logging
import numpy as np
import six
import copy
logger = logging.getLogger(__name__)


class LayerModelHelper(model_helper.ModelHelper):
    """
    Model helper for building models on top of layers abstractions.

    Each layer is the abstraction that is higher level than Operator. Layer
    is responsible for ownership of it's own parameters and can easily be
    instantiated in multiple nets possible with different sets of ops.
    As an example: one can easily instantiate predict and train nets from
    the same set of layers, where predict net will have subset of the
    operators from train net.
    """

    def __init__(self, name, input_feature_schema, trainer_extra_schema,
                 keep_blobs=False):
        ''' TODO(amalevich): more documnetation on input args
        '''

        super(LayerModelHelper, self).__init__(name=name)
        self._layer_names = set()
        self._layers = []
        self._param_to_shape = {}

        # seed default
        self._seed = None
        self._sequence_seed = True

        # optimizer bookkeeping
        self.param_to_optim = {}
        self.param_to_reg = {}

        self._default_optimizer = None
        self._loss = None
        self._prediction = []
        self._output_schema = None

        self._post_grad_net_modifiers = []
        self._final_net_modifiers = []

        # breakdown map; breakdown features are categorical (like dense) but not
        # necessarily used to represent data for training
        self._breakdown_map = None

        # Connect Schema to self.net. That particular instance of schmea will be
        # use for generation of the Layers accross the network and would be used
        # for connection with Readers.
        self._input_feature_schema = schema.NewRecord(
            self.net,
            input_feature_schema
        ) if not keep_blobs else input_feature_schema.clone()
        self._trainer_extra_schema = schema.NewRecord(
            self.net,
            trainer_extra_schema
        ) if not keep_blobs else trainer_extra_schema.clone()
        self._metrics_schema = schema.Struct()

        self._preproc_output_schema = None

        self._init_global_constants()
        self.param_init_net = self.create_init_net('param_init_net')
        self._initialize_params = True

        # additional (hard-coded) diagnose_options to report based on the model
        # TODO(xlwang): it's hack!
        self.ad_hoc_diagnose_blobs_and_operations = []
        self.ad_hoc_plot_blobs = []

    def clear_output_schema(self):
        self._output_schema = None

    def set_initialize_params(self, initialize_params):
        self._initialize_params = initialize_params

    def add_metric_field(self, name, value):
        assert name not in self._metrics_schema.fields, (
            "Try to add metric field twice: {}".format(name))
        self._metrics_schema = self._metrics_schema + schema.Struct(
            (name, value)
        )

    def add_ad_hoc_plot_blob(self, blob, dtype=None):
        assert isinstance(
            blob, (six.string_types, core.BlobReference)
        ), "expect type str or BlobReference, but got {}".format(type(blob))
        dtype = dtype or (np.float, (1, ))
        self.add_metric_field(str(blob), schema.Scalar(dtype, blob))
        self.ad_hoc_plot_blobs.append(blob)

    @staticmethod
    def _get_global_constant_initializer_op(
        blob_name, array=None, dtype=None, initializer=None
    ):
        # to add a global constant to model, one first need to get the
        # initializer
        if array is not None:
            assert initializer is None,\
                "Only one from array and initializer should be specified"
            if dtype is None:
                array = np.array(array)
            else:
                array = np.array(array, dtype=dtype)

            # TODO: make GivenTensor generic
            op_name = None
            if array.dtype == np.int32:
                op_name = 'GivenTensorIntFill'
            elif array.dtype == np.int64:
                op_name = 'GivenTensorInt64Fill'
            elif array.dtype == np.str:
                op_name = 'GivenTensorStringFill'
            elif array.dtype == np.bool:
                op_name = 'GivenTensorBoolFill'
            else:
                op_name = 'GivenTensorFill'

            def initializer(blob_name):
                return core.CreateOperator(
                    op_name, [],
                    blob_name,
                    shape=array.shape,
                    values=array.flatten().tolist()
                )
        else:
            assert initializer is not None
        initializer_op = initializer(blob_name)
        return initializer_op

    def add_global_constant(
        self, name, array=None, dtype=None, initializer=None
    ):
        assert isinstance(name, six.string_types), (
            'name should be a string as we are using it as map key')
        # This is global namescope for constants. They will be created in all
        # init_nets and there should be very few of them.
        assert name not in self.global_constants, \
            "%s already added in global_constants" % name
        blob_name = self.net.NextBlob(name)
        self.global_constants[name] = blob_name
        initializer_op = LayerModelHelper._get_global_constant_initializer_op(
            blob_name, array, dtype, initializer
        )
        assert blob_name not in self.global_constant_initializers, \
            "there is already a initializer op associated with blob %s" % \
            blob_name
        self.global_constant_initializers[blob_name] = initializer_op
        return blob_name

    def maybe_add_global_constant(self, name, *args, **kwargs):
        # To ad hoc add new global constants without duplication
        # if the name was already registered in global_constants, it will not be
        # added even if the intended value is different from its original value

        if name in self.global_constants:
            blob_name = self.global_constants[name]
            initializer_op = \
                LayerModelHelper._get_global_constant_initializer_op(
                    blob_name, *args, **kwargs
                )
            # check if the original initializer is the same as the one intended
            # now
            assert utils.OpAlmostEqual(
                initializer_op,
                self.global_constant_initializers[blob_name],
                'debug_info'
            ), \
                "conflict initializers for global constant %s, " \
                "previous %s, now %s" % (
                    blob_name, str(initializer_op),
                    str(self.global_constant_initializers[blob_name]))
            return blob_name
        return self.add_global_constant(name, *args, **kwargs)

    def _init_global_constants(self):
        self.global_constants = {}
        self.global_constant_initializers = {}
        self.add_global_constant('ONE', 1.0)
        self.add_global_constant('ZERO', 0.0)
        self.add_global_constant('ZERO_RANGE', [0, 0], dtype='int32')

    def _add_global_constants(self, init_net):
        for initializer_op in viewvalues(self.global_constant_initializers):
            init_net._net.op.extend([initializer_op])

    def create_init_net(self, name):
        init_net = core.Net(name)
        self._add_global_constants(init_net)
        return init_net

    def _validate_param_shape(self, param_name, shape):
        if param_name not in self._param_to_shape:
            return

        ref_shape = self._param_to_shape[param_name]

        if shape != ref_shape:
            raise ValueError(
                "Got inconsistent shapes between shared parameters "
                "when trying to map a blob in scope {0} to {1}. ref_shape : "
                " {2}, shape : {3}".format(
                    scope.CurrentNameScope(), param_name, ref_shape, shape)
            )

    def create_param(self, param_name, shape, initializer, optimizer=None,
                     ps_param=None, regularizer=None):
        if isinstance(param_name, core.BlobReference):
            param_name = str(param_name)
        elif isinstance(param_name, six.string_types):
            # Parameter name will be equal to current Namescope that got
            # resolved with the respect of parameter sharing of the scopes.
            param_name = parameter_sharing_context.get_parameter_name(
                param_name)
        else:
            raise ValueError("Unsupported type for param_name")

        param_blob = core.BlobReference(param_name)

        if len(initializer) == 1:
            init_op_args = {}
        else:
            assert len(initializer) == 2
            init_op_args = copy.deepcopy(initializer[1])
        if shape is not None:
            assert 'shape' not in init_op_args
            init_op_args.update({'shape': shape})

        initializer_op = None
        if self._initialize_params:
            initializer_op = core.CreateOperator(
                initializer[0],
                [],
                param_blob,
                **init_op_args
            )

        param = layers.LayerParameter(
            parameter=param_blob,
            initializer=initializer_op,
            optimizer=optimizer,
            ps_param=ps_param,
            regularizer=regularizer
        )

        self._validate_param_shape(param_name, shape)

        self._param_to_shape[param_name] = shape

        return param

    def next_layer_name(self, prefix):
        base_name = core.ScopedName(prefix)
        name = base_name
        index = 0
        while name in self._layer_names:
            name = base_name + '_auto_' + str(index)
            index += 1

        self._layer_names.add(name)
        return name

    def add_layer(self, layer):
        self._layers.append(layer)
        for param in layer.get_parameters():
            assert isinstance(param.parameter, core.BlobReference)

            self.param_to_optim[str(param.parameter)] = \
                param.optimizer or self.default_optimizer

            self.params.append(param.parameter)
            if isinstance(param, layers.LayerParameter):
                self.param_to_reg[param.parameter] = param.regularizer
            elif isinstance(param, ParameterInfo):
                # TODO:
                # Currently, LSTM and RNNcells, which use ModelHelper instead of
                # LayerModelHelper as super class, are called in pooling_methods
                # In ModelHelper, regularization is not supported in create_param
                # We will unify the way of create_param of ModelHelper and
                # LayerModelHelper in the future.
                logger.info('regularization is unsupported for ParameterInfo object')
            else:
                raise ValueError(
                    'unknown object type besides ParameterInfo and LayerParameter: {}'
                    .format(param)
                )

        # The primary value of adding everything to self.net - generation of the
        # operators right away, i.e. if error happens it'll be detected
        # immediately. Other than this - create_x_net should be called.
        layer.add_operators(self.net, self.param_init_net)
        return layer.output_schema

    def get_parameter_blobs(self):
        param_blobs = []
        for layer in self._layers:
            for param in layer.get_parameters():
                param_blobs.append(param.parameter)

        return param_blobs

    def add_post_grad_net_modifiers(self, modifier):
        assert modifier not in self._post_grad_net_modifiers,\
            "{0} is already in {1}".format(modifier, self._post_grad_net_modifiers)
        assert isinstance(modifier, NetModifier),\
            "{} has to be a NetModifier instance".format(modifier)
        self._post_grad_net_modifiers.append(modifier)

    def add_final_net_modifiers(self, modifier):
        assert modifier not in self._final_net_modifiers,\
            "{0} is already in {1}".format(modifier, self._final_net_modifiers)
        assert isinstance(modifier, NetModifier),\
            "{} has to be a NetModifier instance".format(modifier)
        self._final_net_modifiers.append(modifier)

    @property
    def seed(self):
        return self._seed

    @property
    def sequence_seed(self):
        return self._sequence_seed

    def store_seed(self, seed, sequence_seed=True):
        # Store seed config that will be applied to each op in the net.
        self._seed = seed
        # If sequence_seed is True, the i-th op has rand_seed=`seed + i`
        self._sequence_seed = sequence_seed

    def apply_seed(self, net):
        if self._seed:
            net.set_rand_seed(self._seed, self._sequence_seed)

    @property
    def default_optimizer(self):
        return self._default_optimizer

    @default_optimizer.setter
    def default_optimizer(self, optimizer):
        self._default_optimizer = optimizer

    @property
    def input_feature_schema(self):
        return self._input_feature_schema

    @property
    def trainer_extra_schema(self):
        return self._trainer_extra_schema

    @property
    def metrics_schema(self):
        """
        Returns the schema that represents model output that should be used for
        metric reporting.

        During the training/evaluation this schema will be appended to the
        schema that represents model output.
        """
        return self._metrics_schema

    @property
    def output_schema(self):
        assert self._output_schema is not None
        return self._output_schema

    @output_schema.setter
    def output_schema(self, schema):
        assert self._output_schema is None
        self._output_schema = schema

    @property
    def preproc_output_schema(self):
        assert self._preproc_output_schema is not None
        return self._preproc_output_schema

    @preproc_output_schema.setter
    def preproc_output_schema(self, schema):
        assert self._preproc_output_schema is None
        self._preproc_output_schema = schema

    @property
    def prediction(self):
        assert self._prediction, "model prediction is empty"
        return self._prediction

    def add_prediction(self, prediction, weight=1.0):
        assert prediction is not None, "Added prediction should not be None"
        self._prediction.append((prediction, weight))

    @property
    def loss(self):
        assert self._loss is not None
        return self._loss

    @loss.setter
    def loss(self, loss):
        assert self._loss is None
        self._loss = loss

    def has_loss(self):
        return self._loss is not None

    def add_loss(self, loss, name='unnamed'):
        assert loss is not None, "Added loss should not be None"
        assert isinstance(loss, schema.Scalar) or isinstance(
            loss, schema.Struct
        ), "Added loss should be a scalar or a struct"
        if self._loss is None:
            self._loss = schema.Struct((name, loss))
        else:
            # loss could've been set through model.loss directly which could be
            # a scalar
            if isinstance(self._loss, schema.Scalar):
                self._loss = schema.Struct(('unnamed', self._loss))

            prefix_base = name + '_auto_'
            index = 0
            prefix = name
            while prefix in self._loss:
                prefix = prefix_base + str(index)
                index += 1
            loss_struct = schema.Struct((prefix, loss))
            self._loss = self._loss + loss_struct

    def add_output_schema(self, name, value):
        assert value is not None, \
            'Added output schema {} should not be None'.format(name)
        assert isinstance(value, schema.Scalar) or \
            isinstance(value, schema.Struct), \
            'Added output schema {} should be a scalar or a struct.\n\
            Now it is {}.'.format(name, type(value))
        if self._output_schema is None:  # be the first field
            self._output_schema = schema.Struct((name, value))
        else:  # merge with other fields
            assert name not in self._output_schema.fields, \
                'Output Schema Field {} already exists'.format(name)
            self._output_schema = \
                self._output_schema + schema.Struct((name, value))

    def add_trainer_extra_schema(self, trainer_extra_schema):
        trainer_extra_record = schema.NewRecord(self.net, trainer_extra_schema)
        self._trainer_extra_schema += trainer_extra_record

    def __getattr__(self, layer):
        def is_functional_layer(layer):
            if core.IsOperator(layer):
                return True
            elif layer.startswith('FunctionalLayer'):
                return True
            else:
                return False

        def resolve_functional_layer(layer):
            if core.IsOperator(layer):
                return layer
            elif layer.startswith('FunctionalLayer'):
                return layer[len('FunctionalLayer'):]
            else:
                raise ValueError(
                    '%s cannot be resolved as functional layer' % layer
                )

        if layer.startswith('__'):
            raise AttributeError(layer)

        # TODO(amalevich): Add add support for ifbpy inline documentation
        if layers.layer_exists(layer):
            def wrapper(*args, **kwargs):
                new_layer = layers.create_layer(layer, self, *args, **kwargs)
                if kwargs.get("output_to_metrics", False):
                    new_layer.export_output_for_metrics()
                if kwargs.get("params_to_metrics", False):
                    new_layer.export_params_for_metrics()
                return self.add_layer(new_layer)
            return wrapper
        elif is_functional_layer(layer):
            # TODO(xlwang): Desginated layer shadows the usage of an op as a
            # single layer. To enforce using an op (e.g. Split) as functional
            # layer, one can call 'model.FunctionalLayerSplit'
            layer = resolve_functional_layer(layer)

            def wrapper(*args, **kwargs):
                def apply_operator(net, in_record, out_record, **kwargs):
                    # TODO(amalevich): Switch to net.operator as soon as it gets
                    # landed
                    net.__getattr__(layer)(in_record.field_blobs(),
                                           out_record.field_blobs(),
                                           **kwargs)

                if 'name' not in kwargs:
                    kwargs['name'] = layer

                new_layer = layers.create_layer(
                    'Functional',
                    self, *args, function=apply_operator,
                    **kwargs
                )

                if kwargs.get("output_to_metrics", False):
                    new_layer.export_output_for_metrics()
                if kwargs.get("params_to_metrics", False):
                    new_layer.export_params_for_metrics()

                return self.add_layer(new_layer)
            return wrapper
        else:
            # this needs to be an AttributeError to fit hasattr semantics
            raise AttributeError(
                "Trying to create non-registered layer: {}".format(layer))

    @property
    def layers(self):
        return self._layers

    def apply_regularizers_on_loss(
        self,
        train_net,
        train_init_net,
        blob_to_device=None,
    ):
        for param, regularizer in viewitems(self.param_to_reg):
            if regularizer is None:
                continue
            assert isinstance(regularizer, Regularizer)
            added_loss_blob = regularizer(train_net, train_init_net, param, grad=None,
                                          by=RegularizationBy.ON_LOSS)
            if added_loss_blob is not None:
                self.add_loss(
                    schema.Scalar(blob=added_loss_blob),
                    str(added_loss_blob)
                )

    def apply_regularizers_after_optimizer(
        self,
        train_net,
        train_init_net,
        grad_map,
        blob_to_device=None,
    ):
        CPU = muji.OnCPU()
        # if given, blob_to_device is a map from blob to device_option
        blob_to_device = blob_to_device or {}
        for param, regularizer in viewitems(self.param_to_reg):
            if regularizer is None:
                continue
            assert isinstance(regularizer, Regularizer)
            device = get_param_device(
                param,
                grad_map.get(str(param)),
                param_to_device=blob_to_device,
                default_device=CPU,
            )
            with core.DeviceScope(device):
                regularizer(
                    train_net, train_init_net, param, grad=grad_map.get(str(param)),
                    by=RegularizationBy.AFTER_OPTIMIZER
                )

    def apply_post_grad_net_modifiers(
        self,
        trainer_net,
        trainer_init_net,
        grad_map,
        blob_to_device=None,
        modify_output_record=False,
    ):
        param_grad_map = {param: grad_map[param]
                          for param in self.param_to_optim.keys() if param in grad_map}

        for modifier in self._post_grad_net_modifiers:
            modifier(trainer_net, trainer_init_net, param_grad_map,
                     blob_to_device=blob_to_device,
                     modify_output_record=modify_output_record)

    def apply_final_net_modifiers(
        self,
        trainer_net,
        trainer_init_net,
        grad_map,
        blob_to_device=None,
        modify_output_record=False,
    ):
        for modifier in self._final_net_modifiers:
            modifier(trainer_net, trainer_init_net, grad_map,
                     blob_to_device=blob_to_device,
                     modify_output_record=modify_output_record)

    def apply_optimizers(
        self,
        train_net,
        train_init_net,
        grad_map,
        blob_to_device=None,
    ):
        CPU = muji.OnCPU()
        # if given, blob_to_device is a map from blob to device_option
        blob_to_device = blob_to_device or {}
        for param, optimizer in viewitems(self.param_to_optim):
            assert optimizer is not None, \
                "default optimizer must have been set in add_layer"
            # note that not all params has gradient and thus we sent None if
            # gradient does not exists
            device = get_param_device(
                param,
                grad_map.get(str(param)),
                param_to_device=blob_to_device,
                default_device=CPU,
            )
            if device is not None:
                # extra info is not applicable for optimizers
                del device.extra_info[:]

            with core.DeviceScope(device):
                optimizer(
                    train_net, train_init_net, param, grad_map.get(str(param)))

    def _GetOne(self):
        return self.global_constants['ONE']

    # An optimizer which allows us to do NO optimization
    def NoOptim(self, *args, **kwargs):
        pass

    @property
    def breakdown_map(self):
        return self._breakdown_map

    @breakdown_map.setter
    def breakdown_map(self, breakdown_map):
        # TODO(xlwang): provide more rich feature information in breakdown_map;
        # and change the assertion accordingly
        assert isinstance(breakdown_map, dict)
        assert all(isinstance(k, six.string_types) for k in breakdown_map)
        assert sorted(breakdown_map.values()) == list(range(len(breakdown_map)))
        self._breakdown_map = breakdown_map



import math
import sys
import errno
import os
import ctypes
import signal
import torch
import gc
import time
import traceback
import unittest
import subprocess
import itertools
from torch import multiprocessing as mp
from torch.utils.data import Dataset, TensorDataset, DataLoader, ConcatDataset
from torch.utils.data.dataset import random_split
from torch.utils.data.dataloader import default_collate, ExceptionWrapper, MP_STATUS_CHECK_INTERVAL
from common_utils import (TestCase, run_tests, TEST_NUMPY, IS_WINDOWS, IS_PPC, NO_MULTIPROCESSING_SPAWN,
                          skipIfRocm, load_tests)

# load_tests from common_utils is used to automatically filter tests for
# sharding on sandcastle. This line silences flake warnings
load_tests = load_tests

# We cannot import TEST_CUDA from common_cuda here, because if we do that,
# the TEST_CUDNN line from common_cuda will be executed multiple times
# as well during the execution of this test suite, and it will cause
# CUDA OOM error on Windows.
TEST_CUDA = torch.cuda.is_available()

if not NO_MULTIPROCESSING_SPAWN:
    # Get a multiprocessing context because some test / third party library will
    # set start_method when imported, and setting again triggers RuntimeError.
    mp = mp.get_context(method='spawn')


JOIN_TIMEOUT = 17.0 if IS_WINDOWS or IS_PPC else 8.5


class TestDatasetRandomSplit(TestCase):
    def test_lengths_must_equal_datset_size(self):
        with self.assertRaises(ValueError):
            random_split([1, 2, 3, 4], [1, 2])

    def test_splits_have_correct_size(self):
        splits = random_split([1, 2, 3, 4, 5, 6], [2, 4])
        self.assertEqual(len(splits), 2)
        self.assertEqual(len(splits[0]), 2)
        self.assertEqual(len(splits[1]), 4)

    def test_splits_are_mutually_exclusive(self):
        data = [5, 2, 3, 4, 1, 6]
        splits = random_split(data, [2, 4])
        all_values = []
        all_values.extend(list(splits[0]))
        all_values.extend(list(splits[1]))
        data.sort()
        all_values.sort()
        self.assertListEqual(data, all_values)


class TestTensorDataset(TestCase):

    def test_len(self):
        source = TensorDataset(torch.randn(15, 10, 2, 3, 4, 5), torch.randperm(15))
        self.assertEqual(len(source), 15)

    def test_getitem(self):
        t = torch.randn(15, 10, 2, 3, 4, 5)
        l = torch.randn(15, 10)
        source = TensorDataset(t, l)
        for i in range(15):
            self.assertEqual(t[i], source[i][0])
            self.assertEqual(l[i], source[i][1])

    def test_getitem_1d(self):
        t = torch.randn(15)
        l = torch.randn(15)
        source = TensorDataset(t, l)
        for i in range(15):
            self.assertEqual(t[i], source[i][0])
            self.assertEqual(l[i], source[i][1])

    def test_single_tensor(self):
        t = torch.randn(5, 10)
        source = TensorDataset(t)
        self.assertEqual(len(source), 5)
        for i in range(5):
            self.assertEqual(t[i], source[i][0])

    def test_many_tensors(self):
        t0 = torch.randn(5, 10, 2, 3, 4, 5)
        t1 = torch.randn(5, 10)
        t2 = torch.randn(5, 10, 2, 5)
        t3 = torch.randn(5, 10, 3, 7)
        source = TensorDataset(t0, t1, t2, t3)
        self.assertEqual(len(source), 5)
        for i in range(5):
            self.assertEqual(t0[i], source[i][0])
            self.assertEqual(t1[i], source[i][1])
            self.assertEqual(t2[i], source[i][2])
            self.assertEqual(t3[i], source[i][3])


class TestConcatDataset(TestCase):

    def test_concat_two_singletons(self):
        result = ConcatDataset([[0], [1]])
        self.assertEqual(2, len(result))
        self.assertEqual(0, result[0])
        self.assertEqual(1, result[1])

    def test_concat_two_non_singletons(self):
        result = ConcatDataset([[0, 1, 2, 3, 4],
                                [5, 6, 7, 8, 9]])
        self.assertEqual(10, len(result))
        self.assertEqual(0, result[0])
        self.assertEqual(5, result[5])

    def test_concat_two_non_singletons_with_empty(self):
        # Adding an empty dataset somewhere is correctly handled
        result = ConcatDataset([[0, 1, 2, 3, 4],
                                [],
                                [5, 6, 7, 8, 9]])
        self.assertEqual(10, len(result))
        self.assertEqual(0, result[0])
        self.assertEqual(5, result[5])

    def test_concat_raises_index_error(self):
        result = ConcatDataset([[0, 1, 2, 3, 4],
                                [5, 6, 7, 8, 9]])
        with self.assertRaises(IndexError):
            # this one goes to 11
            result[11]

    def test_add_dataset(self):
        d1 = TensorDataset(torch.rand(7, 3, 28, 28), torch.rand(7))
        d2 = TensorDataset(torch.rand(7, 3, 28, 28), torch.rand(7))
        d3 = TensorDataset(torch.rand(7, 3, 28, 28), torch.rand(7))
        result = d1 + d2 + d3
        self.assertEqual(21, len(result))
        self.assertEqual(0, (d1[0][0] - result[0][0]).abs().sum())
        self.assertEqual(0, (d2[0][0] - result[7][0]).abs().sum())
        self.assertEqual(0, (d3[0][0] - result[14][0]).abs().sum())


# Stores the first encountered exception in .exception.
# Inspired by https://stackoverflow.com/a/33599967
class ErrorTrackingProcess(mp.Process):

    def __init__(self, *args, **kwargs):
        super(ErrorTrackingProcess, self).__init__(*args, **kwargs)
        self._pconn, self._cconn = mp.Pipe()
        self._exception = None

    def run(self):
        # Disable polluting stderr with errors that are supposed to happen.
        sys.stderr = open(os.devnull, "w")
        try:
            super(ErrorTrackingProcess, self).run()
            self._cconn.send(None)
        except Exception:
            self._cconn.send(ExceptionWrapper(sys.exc_info()))
            raise

    @property
    def exception(self):
        if self._pconn.poll():
            self._exception = self._pconn.recv()
        if self._exception is None:
            return None
        else:
            return self._exception.exc_type(self._exception.exc_msg)

    # ESRCH means that os.kill can't finds alive proc
    def send_signal(self, signum, ignore_ESRCH=False):
        try:
            os.kill(self.pid, signum)
        except OSError as e:
            if not ignore_ESRCH or e.errno != errno.ESRCH:
                raise


class ErrorDataset(Dataset):

    def __init__(self, size):
        self.size = size

    def __len__(self):
        return self.size


class SegfaultDataset(Dataset):

    def __init__(self, size):
        self.size = size

    def __getitem__(self, idx):
        return ctypes.string_at(0)

    def __len__(self):
        return self.size


class SleepDataset(Dataset):

    def __init__(self, size, sleep_sec):
        self.size = size
        self.sleep_sec = sleep_sec
        self.sleeped = False

    def __getitem__(self, idx):
        if not self.sleeped:
            time.sleep(self.sleep_sec)
            self.sleeped = True
        return idx

    def __len__(self):
        return self.size


class SeedDataset(Dataset):

    def __init__(self, size):
        self.size = size

    def __getitem__(self, idx):
        return torch.initial_seed()

    def __len__(self):
        return self.size


# Inspired by https://stackoverflow.com/a/26703365
# This will ensure that each worker at least processes one data
class SynchronizedSeedDataset(Dataset):

    def __init__(self, size, num_workers):
        assert size >= num_workers
        self.count = mp.Value('i', 0, lock=True)
        self.barrier = mp.Semaphore(0)
        self.num_workers = num_workers
        self.size = size

    def __getitem__(self, idx):
        with self.count.get_lock():
            self.count.value += 1
            if self.count.value == self.num_workers:
                self.barrier.release()
        self.barrier.acquire()
        self.barrier.release()
        return torch.initial_seed()

    def __len__(self):
        return self.size


def _test_timeout():
    dataset = SleepDataset(10, 3)
    dataloader = DataLoader(dataset, batch_size=2, num_workers=2, timeout=1)
    _ = next(iter(dataloader))


def _test_timeout_pin_memory():
    dataset = SleepDataset(10, 3)
    dataloader = DataLoader(dataset, batch_size=2, num_workers=2, timeout=1, pin_memory=True)
    _ = next(iter(dataloader))


def disable_stderr(worker_id):
    r"""
    Avoids printing "ERROR: Unexpected segmentation fault encountered in worker."
    from workers. Since worker signal handler prints with low-level write(),
    this has to be done on OS level via dup.

    This is used as worker_init_fn for test_segfault.
    """
    sys.stderr.flush()  # flush library buffers that dup2 knows nothing about
    # Can't use a with-block because otherwise the fd will be closed when this
    # function ends.
    devnull = open(os.devnull, 'w')
    os.dup2(devnull.fileno(), sys.stderr.fileno())


def _test_segfault():
    dataset = SegfaultDataset(10)
    dataloader = DataLoader(dataset, batch_size=2, num_workers=2, worker_init_fn=disable_stderr)
    _ = next(iter(dataloader))


class TestProperExitDataset(object):
    def __init__(self, size, error_event):
        self.size = size
        self.error_event = error_event

    def __len__(self):
        return self.size

    def __getitem__(self, idx):
        if self.error_event is not None and self.error_event.is_set():
            raise RuntimeError('Worker error')
        return torch.tensor([idx])


# See TestDataLoader.test_proper_exit for usage
def _test_proper_exit(use_workers, pin_memory, exit_method, hold_iter_reference,
                      worker_pids, setup_event):
    num_workers = 2 if use_workers else 0

    if exit_method == 'worker_error' or exit_method == 'worker_kill':
        assert use_workers is True

    ds = TestProperExitDataset(10, setup_event if exit_method == 'worker_error' else None)

    loader = DataLoader(ds, batch_size=2, shuffle=False,
                        num_workers=num_workers, pin_memory=pin_memory)
    error_it = 4
    assert len(loader) > error_it

    it = iter(loader)
    if use_workers:
        for i, w in enumerate(it.workers):
            worker_pids[i] = w.pid

    def kill_pid(pid):
        if IS_WINDOWS:
            os.system('taskkill /PID ' + str(os.getpid()) + ' /F')
        else:
            os.kill(os.getpid(), signal.SIGKILL)

    for i, _ in enumerate(it):
        if i == 0:
            if not hold_iter_reference:
                del it
            setup_event.set()
        if i == error_it:
            if exit_method == 'main_error':
                raise RuntimeError('Error')
            elif exit_method == 'main_kill':
                kill_pid(os.getpid())
            elif exit_method == 'worker_kill':
                kill_pid(worker_pids[0])

    if not hold_iter_reference:
        # Tries to trigger the __del__ clean-up rather than the automatic
        # exiting of daemonic children. Technically it should be automatically
        # triggered, but I don't want to rely on the implementation detail of
        # Python gc.
        gc.collect()


# test custom init function
def init_fn(worker_id):
    torch.manual_seed(12345)


class TestDataLoader(TestCase):

    def setUp(self):
        self.data = torch.randn(100, 2, 3, 5)
        self.labels = torch.randperm(50).repeat(2)
        self.dataset = TensorDataset(self.data, self.labels)

    def _test_sequential(self, loader):
        batch_size = loader.batch_size
        for i, (sample, target) in enumerate(loader):
            idx = i * batch_size
            self.assertEqual(sample, self.data[idx:idx + batch_size])
            self.assertEqual(target, self.labels[idx:idx + batch_size])
        self.assertEqual(i, math.floor((len(self.dataset) - 1) / batch_size))

    def _test_shuffle(self, loader):
        found_data = {i: 0 for i in range(self.data.size(0))}
        found_labels = {i: 0 for i in range(self.labels.size(0))}
        batch_size = loader.batch_size
        for i, (batch_samples, batch_targets) in enumerate(loader):
            for sample, target in zip(batch_samples, batch_targets):
                for data_point_idx, data_point in enumerate(self.data):
                    if data_point.eq(sample).all():
                        self.assertFalse(found_data[data_point_idx])
                        found_data[data_point_idx] += 1
                        break
                self.assertEqual(target, self.labels[data_point_idx])
                found_labels[data_point_idx] += 1
            self.assertEqual(sum(found_data.values()), (i + 1) * batch_size)
            self.assertEqual(sum(found_labels.values()), (i + 1) * batch_size)
        self.assertEqual(i, math.floor((len(self.dataset) - 1) / batch_size))

    def _test_error(self, loader):
        it = iter(loader)
        errors = 0
        while True:
            try:
                next(it)
            except NotImplementedError:
                errors += 1
            except StopIteration:
                self.assertEqual(errors,
                                 math.ceil(float(len(loader.dataset)) / loader.batch_size))
                return

    def test_invalid_assign_after_init(self):
        dl = DataLoader(self.dataset)
        for attr in ('batch_size', 'sampler', 'drop_last'):
            def fn():
                setattr(dl, attr, {})

            self.assertRaises(ValueError, fn)

    def test_sequential(self):
        self._test_sequential(DataLoader(self.dataset))

    def test_sequential_batch(self):
        self._test_sequential(DataLoader(self.dataset, batch_size=2))

    def test_growing_dataset(self):
        dataset = [torch.ones(4) for _ in range(4)]
        dataloader_seq = DataLoader(dataset, shuffle=False)
        dataloader_shuffle = DataLoader(dataset, shuffle=True)
        dataset.append(torch.ones(4))
        self.assertEqual(len(dataloader_seq), 5)
        self.assertEqual(len(dataloader_shuffle), 5)

    @unittest.skipIf(not TEST_CUDA, "CUDA unavailable")
    @skipIfRocm
    def test_sequential_pin_memory(self):
        loader = DataLoader(self.dataset, batch_size=2, pin_memory=True)
        for input, target in loader:
            self.assertTrue(input.is_pinned())
            self.assertTrue(target.is_pinned())

    def test_multiple_dataloaders(self):
        loader1_it = iter(DataLoader(self.dataset, num_workers=1))
        loader2_it = iter(DataLoader(self.dataset, num_workers=2))
        next(loader1_it)
        next(loader1_it)
        next(loader2_it)
        next(loader2_it)
        next(loader1_it)
        next(loader2_it)

    @unittest.skip("temporarily disable until flaky failures are fixed")
    def test_segfault(self):
        p = ErrorTrackingProcess(target=_test_segfault)
        p.start()
        p.join(JOIN_TIMEOUT)
        try:
            self.assertFalse(p.is_alive())
            self.assertNotEqual(p.exitcode, 0)
            if IS_WINDOWS:
                self.assertIsInstance(p.exception, OSError)
                self.assertRegex(str(p.exception), r'access violation reading ')
            else:
                self.assertIsInstance(p.exception, RuntimeError)
                self.assertRegex(str(p.exception), r'DataLoader worker \(pid \d+\) is killed by signal: ')
        finally:
            p.terminate()

    @skipIfRocm
    def test_timeout(self):
        if TEST_CUDA and not NO_MULTIPROCESSING_SPAWN:
            targets = (_test_timeout, _test_timeout_pin_memory)
        else:
            targets = (_test_timeout,)
        for target in targets:
            p = ErrorTrackingProcess(target=target)
            p.start()
            p.join(JOIN_TIMEOUT)
            try:
                self.assertFalse(p.is_alive())
                self.assertNotEqual(p.exitcode, 0)
                self.assertIsInstance(p.exception, RuntimeError)
                self.assertRegex(str(p.exception), r'DataLoader timed out after \d+ seconds')
            finally:
                p.terminate()

    def test_worker_seed(self):
        num_workers = 6
        dataset = SynchronizedSeedDataset(num_workers, num_workers)
        dataloader = DataLoader(dataset, batch_size=1, num_workers=num_workers)
        seeds = set()
        for batch in dataloader:
            seeds.add(batch[0])
        self.assertEqual(len(seeds), num_workers)

    def test_worker_init_fn(self):
        dataset = SeedDataset(4)
        dataloader = DataLoader(dataset, batch_size=2, num_workers=2,
                                worker_init_fn=init_fn)
        for batch in dataloader:
            self.assertEqual(12345, batch[0])
            self.assertEqual(12345, batch[1])

    def test_shuffle(self):
        self._test_shuffle(DataLoader(self.dataset, shuffle=True))

    def test_shuffle_batch(self):
        self._test_shuffle(DataLoader(self.dataset, batch_size=2, shuffle=True))

    def test_sequential_workers(self):
        self._test_sequential(DataLoader(self.dataset, num_workers=4))

    def test_seqential_batch_workers(self):
        self._test_sequential(DataLoader(self.dataset, batch_size=2, num_workers=4))

    def test_shuffle_workers(self):
        self._test_shuffle(DataLoader(self.dataset, shuffle=True, num_workers=4))

    def test_shuffle_batch_workers(self):
        self._test_shuffle(DataLoader(self.dataset, batch_size=2, shuffle=True, num_workers=4))

    def _test_batch_sampler(self, **kwargs):
        # [(0, 1), (2, 3, 4), (5, 6), (7, 8, 9), ...]
        batches = []
        for i in range(0, 100, 5):
            batches.append(tuple(range(i, i + 2)))
            batches.append(tuple(range(i + 2, i + 5)))

        dl = DataLoader(self.dataset, batch_sampler=batches, **kwargs)
        self.assertEqual(len(dl), 40)
        for i, (input, _target) in enumerate(dl):
            if i % 2 == 0:
                offset = i * 5 // 2
                self.assertEqual(len(input), 2)
                self.assertEqual(input, self.data[offset:offset + 2])
            else:
                offset = i * 5 // 2
                self.assertEqual(len(input), 3)
                self.assertEqual(input, self.data[offset:offset + 3])

    def test_RandomSampler(self):

        from collections import Counter
        from torch.utils.data import RandomSampler

        def sample_stat(sampler, num_samples):
            counts = Counter(sampler)
            count_repeated = sum(val > 1 for val in counts.values())
            return (count_repeated, min(counts.keys()), max(counts.keys()))

        # test sample with replacement
        n = len(self.dataset) + 1  # ensure at least one sample is drawn more than once
        sampler_with_replacement = RandomSampler(self.dataset, replacement=True, num_samples=n)
        count_repeated, minval, maxval = sample_stat(sampler_with_replacement, n)
        self.assertTrue(count_repeated > 0)
        self.assertTrue(minval >= 0)
        self.assertTrue(maxval < len(self.dataset))

        # test sample without replacement
        sampler_without_replacement = RandomSampler(self.dataset)
        count_repeated, minval, maxval = sample_stat(sampler_without_replacement, len(self.dataset))
        self.assertTrue(count_repeated == 0)
        self.assertTrue(minval == 0)
        self.assertTrue(maxval == len(self.dataset) - 1)

        # raise error when replacement=False and num_samples is not None
        self.assertRaises(ValueError, lambda: RandomSampler(self.dataset, num_samples=len(self.dataset)))

        self.assertRaises(ValueError, lambda: RandomSampler(self.dataset, num_samples=0))

    def test_duplicating_data_with_drop_last(self):

        from torch.utils.data.distributed import DistributedSampler

        num_processes = 4
        num_batches = 9
        data_set = torch.IntTensor(range(num_batches))
        scanned_data = torch.IntTensor([])
        for i in range(num_processes):
            s = DistributedSampler(data_set, num_processes, i)
            d_loader = DataLoader(data_set, batch_size=int(num_batches / num_processes), drop_last=True, sampler=s)
            for k, data in enumerate(d_loader):
                scanned_data = torch.cat((scanned_data, data), 0)

        self.assertEqual(scanned_data.size(), scanned_data.unique().size())

    @unittest.skipIf(NO_MULTIPROCESSING_SPAWN, "Disabled for environments that \
                     don't support multiprocessing with spawn start method")
    def test_batch_sampler(self):
        self._test_batch_sampler()
        self._test_batch_sampler(num_workers=4)

    @unittest.skipIf(not TEST_CUDA, "CUDA unavailable")
    @skipIfRocm
    def test_shuffle_pin_memory(self):
        loader = DataLoader(self.dataset, batch_size=2, shuffle=True, num_workers=4, pin_memory=True)
        for input, target in loader:
            self.assertTrue(input.is_pinned())
            self.assertTrue(target.is_pinned())

    @unittest.skipIf(not TEST_NUMPY, "numpy unavailable")
    def test_numpy(self):
        import numpy as np

        class TestDataset(torch.utils.data.Dataset):
            def __getitem__(self, i):
                return np.ones((2, 3, 4)) * i

            def __len__(self):
                return 1000

        loader = DataLoader(TestDataset(), batch_size=12)
        batch = next(iter(loader))
        self.assertIsInstance(batch, torch.DoubleTensor)
        self.assertEqual(batch.size(), torch.Size([12, 2, 3, 4]))

    def test_error(self):
        self._test_error(DataLoader(ErrorDataset(100), batch_size=2, shuffle=True))

    @unittest.skipIf(NO_MULTIPROCESSING_SPAWN, "Disabled for environments that \
                     don't support multiprocessing with spawn start method")
    def test_error_workers(self):
        self._test_error(DataLoader(ErrorDataset(41), batch_size=2, shuffle=True, num_workers=4))

    @unittest.skipIf(IS_WINDOWS, "FIXME: stuck test")
    def test_partial_workers(self):
        r"""Check that workers exit even if the iterator is not exhausted."""
        if TEST_CUDA:
            pin_memory_configs = (True, False)
        else:
            pin_memory_configs = (False,)

        for pin_memory in pin_memory_configs:
            loader = iter(DataLoader(self.dataset, batch_size=2, num_workers=4, pin_memory=pin_memory))
            workers = loader.workers
            if pin_memory:
                pin_memory_thread = loader.pin_memory_thread
            for i, sample in enumerate(loader):
                if i == 10:
                    break
            assert i == 10
            del loader
            for w in workers:
                w.join(JOIN_TIMEOUT)
                self.assertFalse(w.is_alive(), 'subprocess not terminated')
            if pin_memory:
                pin_memory_thread.join(JOIN_TIMEOUT)
                self.assertFalse(pin_memory_thread.is_alive())

    @staticmethod
    def _is_process_alive(pid, pname):
        # There is a chance of a terminated child process's pid being reused by a new unrelated process,
        # but since we are looping this check very frequently, we will know that the child process dies
        # before the new unrelated process starts.
        if IS_WINDOWS:
            command = 'tasklist | find "{}" /i'.format(pid)
        else:
            command = 'ps -p {} -o comm='.format(pid)
        p = subprocess.Popen(command, stdout=subprocess.PIPE, shell=True)
        (output, err) = p.communicate()
        p_status = p.wait()
        output = output.decode('utf-8')
        return pname in output

    @skipIfRocm
    def test_proper_exit(self):
        (r'''There might be ConnectionResetError or leaked semaphore warning '''
         r'''(due to dirty process exit), but they are all safe to ignore''')

        # TODO: test the case where the pin_memory_thread triggers an
        #       error/fatal signal. I haven't found out how to properly do that.

        # Array to store the worker pids.
        worker_pids = mp.Array('i', [-1 for _ in range(10)])

        def wait_pids(pids, timeout):
            r"""Wait for all process specified in pids to exit in given timeout."""
            exit_status = [False for _ in pids]
            start_time = time.time()
            pname = 'python'
            while True:
                for i in range(len(pids)):
                    pid = pids[i]
                    if not exit_status[i]:
                        if not TestDataLoader._is_process_alive(pid, pname):
                            exit_status[i] = True
                if all(exit_status):
                    break
                else:
                    if time.time() - start_time > timeout:
                        break
                    time.sleep(0.5)
            return exit_status

        for use_workers, pin_memory, hold_iter_reference in itertools.product([True, False], repeat=3):
            # `hold_iter_reference` specifies whether we hold a reference to the
            # iterator. This is interesting because Python3 error traces holds a
            # reference to the frames, which hold references to all the local
            # variables including the iterator, and then the iterator dtor may
            # not be called before process end. It is important to see that the
            # processes still exit in both cases.

            if pin_memory and (not TEST_CUDA or NO_MULTIPROCESSING_SPAWN):
                # Can't use CUDA without spawn
                continue

            # `exit_method` controls the way the loader process ends.
            #   - `*_kill` means that `*` is killed by OS.
            #   - `*_error` means that `*` raises an error.
            #   - `None` means that no error happens.
            # In all cases, all processes should end properly.
            if use_workers:
                exit_methods = [None, 'main_error', 'main_kill', 'worker_kill', 'worker_error']
            else:
                exit_methods = [None, 'main_error', 'main_kill']

            for exit_method in exit_methods:

                # clear pids array first
                for i in range(len(worker_pids)):
                    worker_pids[i] = -1

                # Event that the loader process uses to signal testing process
                # that various things are setup, including that the worker pids
                # are specified in `worker_pids` array.
                setup_event = mp.Event()

                p = ErrorTrackingProcess(target=_test_proper_exit,
                                         args=(use_workers, pin_memory, exit_method,
                                               hold_iter_reference, worker_pids, setup_event))
                p.start()

                # Wait for loader process to set everything up, i.e., filling
                # worker pids in `worker_pids`.
                setup_event.wait(timeout=JOIN_TIMEOUT)
                self.assertTrue(setup_event.is_set(), 'loader process setup timed out')

                pids = [pid for pid in worker_pids if pid > 0]

                try:
                    exit_status = wait_pids(pids, timeout=(MP_STATUS_CHECK_INTERVAL + JOIN_TIMEOUT))
                    if not all(exit_status):
                        self.fail('subprocess (pid(s) {}) not terminated'.format(
                            ', '.join(p for p, exited in zip(pids, exit_status) if not exited)))
                    p.join(JOIN_TIMEOUT + MP_STATUS_CHECK_INTERVAL)
                    self.assertFalse(p.is_alive(), 'loader process not terminated')
                    if exit_method is None:
                        self.assertEqual(p.exitcode, 0)
                    else:
                        self.assertNotEqual(p.exitcode, 0)
                finally:
                    p.terminate()

    def test_len(self):
        def check_len(dl, expected):
            self.assertEqual(len(dl), expected)
            n = 0
            for sample in dl:
                n += 1
            self.assertEqual(n, expected)
        check_len(self.dataset, 100)
        check_len(DataLoader(self.dataset, batch_size=2), 50)
        check_len(DataLoader(self.dataset, batch_size=3), 34)

    @unittest.skipIf(not TEST_NUMPY, "numpy unavailable")
    def test_numpy_scalars(self):
        import numpy as np

        class ScalarDataset(torch.utils.data.Dataset):
            def __init__(self, dtype):
                self.dtype = dtype

            def __getitem__(self, i):
                return self.dtype()

            def __len__(self):
                return 4

        dtypes = {
            np.float64: torch.DoubleTensor,
            np.float32: torch.FloatTensor,
            np.float16: torch.HalfTensor,
            np.int64: torch.LongTensor,
            np.int32: torch.IntTensor,
            np.int16: torch.ShortTensor,
            np.int8: torch.CharTensor,
            np.uint8: torch.ByteTensor,
        }
        for dt, tt in dtypes.items():
            dset = ScalarDataset(dt)
            loader = DataLoader(dset, batch_size=2)
            batch = next(iter(loader))
            self.assertIsInstance(batch, tt)

    @unittest.skipIf(not TEST_NUMPY, "numpy unavailable")
    def test_default_collate_bad_numpy_types(self):
        import numpy as np

        # Should be a no-op
        arr = np.array(['a', 'b', 'c'])
        default_collate(arr)

        arr = np.array([[['a', 'b', 'c']]])
        self.assertRaises(TypeError, lambda: default_collate(arr))

        arr = np.array([object(), object(), object()])
        self.assertRaises(TypeError, lambda: default_collate(arr))

        arr = np.array([[[object(), object(), object()]]])
        self.assertRaises(TypeError, lambda: default_collate(arr))


class StringDataset(Dataset):
    def __init__(self):
        self.s = '12345'

    def __len__(self):
        return len(self.s)

    def __getitem__(self, ndx):
        return (self.s[ndx], ndx)


class TestStringDataLoader(TestCase):
    def setUp(self):
        self.dataset = StringDataset()

    @unittest.skipIf(not TEST_CUDA, "CUDA unavailable")
    @skipIfRocm
    def test_shuffle_pin_memory(self):
        loader = DataLoader(self.dataset, batch_size=2, shuffle=True, num_workers=4, pin_memory=True)
        for batch_ndx, (s, n) in enumerate(loader):
            self.assertIsInstance(s[0], str)
            self.assertTrue(n.is_pinned())


class DictDataset(Dataset):
    def __len__(self):
        return 4

    def __getitem__(self, ndx):
        return {
            'a_tensor': torch.Tensor(4, 2).fill_(ndx),
            'another_dict': {
                'a_number': ndx,
            },
        }


class TestDictDataLoader(TestCase):
    def setUp(self):
        self.dataset = DictDataset()

    def test_sequential_batch(self):
        loader = DataLoader(self.dataset, batch_size=2, shuffle=False)
        batch_size = loader.batch_size
        for i, sample in enumerate(loader):
            idx = i * batch_size
            self.assertEqual(set(sample.keys()), {'a_tensor', 'another_dict'})
            self.assertEqual(set(sample['another_dict'].keys()), {'a_number'})

            t = sample['a_tensor']
            self.assertEqual(t.size(), torch.Size([batch_size, 4, 2]))
            self.assertTrue((t[0] == idx).all())
            self.assertTrue((t[1] == idx + 1).all())

            n = sample['another_dict']['a_number']
            self.assertEqual(n.size(), torch.Size([batch_size]))
            self.assertEqual(n[0], idx)
            self.assertEqual(n[1], idx + 1)

    @unittest.skipIf(not TEST_CUDA, "CUDA unavailable")
    @skipIfRocm
    def test_pin_memory(self):
        loader = DataLoader(self.dataset, batch_size=2, pin_memory=True)
        for batch_ndx, sample in enumerate(loader):
            self.assertTrue(sample['a_tensor'].is_pinned())
            self.assertTrue(sample['another_dict']['a_number'].is_pinned())


class TestWorkerQueueDataset(Dataset):
    def __init__(self, data):
        self.data = data
        self.worker_id = None

    def worker_init_fn(self, worker_id):
        self.worker_id = worker_id

    def __getitem__(self, item):
        return self.worker_id, self.data[item]

    def __len__(self):
        return len(self.data)


class TestIndividualWorkerQueue(TestCase):
    def setUp(self):
        self.dataset = TestWorkerQueueDataset([i for i in range(128)])

    def _run_ind_worker_queue_test(self, batch_size, num_workers):
        loader = DataLoader(
            self.dataset, batch_size=batch_size, shuffle=False, num_workers=num_workers,
            worker_init_fn=self.dataset.worker_init_fn
        )
        current_worker_idx = 0
        for i, (worker_ids, sample) in enumerate(loader):
            self.assertEqual(worker_ids.tolist(), [current_worker_idx] * batch_size)
            self.assertEqual(sample.tolist(), [j for j in range(i * batch_size, (i + 1) * batch_size)])
            current_worker_idx += 1
            if current_worker_idx == num_workers:
                current_worker_idx = 0

    def test_ind_worker_queue(self):
        for batch_size in (8, 16, 32, 64):
            for num_workers in range(1, 6):
                self._run_ind_worker_queue_test(batch_size=batch_size, num_workers=num_workers)


if __name__ == '__main__':
    run_tests()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import functools
import os
import unittest
import sys
import torch
import torch.autograd.function as function

pytorch_test_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
sys.path.insert(-1, pytorch_test_dir)

from common_utils import *

torch.set_default_tensor_type('torch.FloatTensor')


def _skipper(condition, reason):
    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            if condition():
                raise unittest.SkipTest(reason)
            return f(*args, **kwargs)
        return wrapper
    return decorator


skipIfNoCuda = _skipper(lambda: not torch.cuda.is_available(),
                        'CUDA is not available')

skipIfTravis = _skipper(lambda: os.getenv('TRAVIS'),
                        'Skip In Travis')

skipIfCI = _skipper(lambda: os.getenv('CI') or os.getenv('TRAVIS') or
                    os.getenv('JENKINS_URL') or os.getenv('INTEGRATED'),
                    'Skip In CI')


def flatten(x):
    return tuple(function._iter_filter(lambda o: isinstance(o, torch.Tensor))(x))



## @package net_printer
# Module caffe2.python.net_printer
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from caffe2.proto.caffe2_pb2 import OperatorDef, NetDef
from caffe2.python.checkpoint import Job
from caffe2.python.core import Net, ExecutionStep, Plan
from caffe2.python.task import Task, TaskGroup, WorkspaceType, TaskOutput
from collections import defaultdict
from contextlib import contextmanager
from copy import copy
from future.utils import viewkeys
from itertools import chain
from six import binary_type, text_type


class Visitor(object):
    @classmethod
    def register(cls, Type):
        if not(hasattr(cls, 'visitors')):
            cls.visitors = {}
        else:
            assert Type not in cls.visitors, \
                '{} already registered!'.format(Type)

        def _register(func):
            cls.visitors[Type] = func
            return func

        return _register

    def __call__(self, obj, *args, **kwargs):
        if obj is None:
            return

        Type = type(obj)
        if Type not in self.__class__.visitors:
            raise TypeError('%s: unsupported object type: %s' % (
                self.__class__.__name__, Type))

        func = self.__class__.visitors[Type]
        return func(self, obj, *args, **kwargs)


class Analyzer(Visitor):
    PREFIXES_TO_IGNORE = {'distributed_ctx_init'}

    def __init__(self):
        self.workspaces = defaultdict(lambda: defaultdict(lambda: 0))
        self.workspace_ctx = []

    @property
    def workspace(self):
        return self.workspace_ctx[-1]

    @contextmanager
    def set_workspace(self, node=None, ws=None, do_copy=False):
        if ws is not None:
            ws = ws
        elif node is not None:
            ws = self.workspaces[str(node)]
        else:
            ws = self.workspace
        if do_copy:
            ws = copy(ws)
        self.workspace_ctx.append(ws)
        yield ws
        del self.workspace_ctx[-1]

    def define_blob(self, blob):
        self.workspace[blob] += 1

    def need_blob(self, blob):
        if any(blob.startswith(p) for p in Analyzer.PREFIXES_TO_IGNORE):
            return
        assert blob in self.workspace, 'Blob undefined: %s' % blob


@Analyzer.register(OperatorDef)
def analyze_op(analyzer, op):
    for x in op.input:
        analyzer.need_blob(x)
    for x in op.output:
        analyzer.define_blob(x)


@Analyzer.register(Net)
def analyze_net(analyzer, net):
    for x in net.Proto().op:
        analyzer(x)


@Analyzer.register(ExecutionStep)
def analyze_step(analyzer, step):
    proto = step.Proto()
    with analyzer.set_workspace(do_copy=proto.create_workspace):
        if proto.report_net:
            with analyzer.set_workspace(do_copy=True):
                analyzer(step.get_net(proto.report_net))
        all_new_blobs = set()
        substeps = step.Substeps() + [step.get_net(n) for n in proto.network]
        for substep in substeps:
            with analyzer.set_workspace(
                    do_copy=proto.concurrent_substeps) as ws_in:
                analyzer(substep)
                if proto.should_stop_blob:
                    analyzer.need_blob(proto.should_stop_blob)
            if proto.concurrent_substeps:
                new_blobs = set(viewkeys(ws_in)) - set(viewkeys(analyzer.workspace))
                assert len(all_new_blobs & new_blobs) == 0, (
                    'Error: Blobs created by multiple parallel steps: %s' % (
                        ', '.join(all_new_blobs & new_blobs)))
                all_new_blobs |= new_blobs
    for x in all_new_blobs:
        analyzer.define_blob(x)


@Analyzer.register(Task)
def analyze_task(analyzer, task):
    # check that our plan protobuf is not too large (limit of 64Mb)
    step = task.get_step()
    plan = Plan(task.node)
    plan.AddStep(step)
    proto_len = len(plan.Proto().SerializeToString())
    assert proto_len < 2 ** 26, (
        'Due to a protobuf limitation, serialized tasks must be smaller '
        'than 64Mb, but this task has {} bytes.' % proto_len)

    is_private = task.workspace_type() != WorkspaceType.GLOBAL
    with analyzer.set_workspace(do_copy=is_private):
        analyzer(step)


@Analyzer.register(TaskGroup)
def analyze_task_group(analyzer, tg):
    for task in tg.tasks_by_node().tasks():
        with analyzer.set_workspace(node=task.node):
            analyzer(task)


@Analyzer.register(Job)
def analyze_job(analyzer, job):
    analyzer(job.init_group)
    analyzer(job.epoch_group)


def analyze(obj):
    """
    Given a Job, visits all the execution steps making sure that:
      - no undefined blobs will be found during excution
      - no blob with same name is defined in concurrent steps
    """
    Analyzer()(obj)


class Text(object):
    def __init__(self):
        self._indent = 0
        self._lines_in_context = [0]
        self.lines = []

    @contextmanager
    def context(self, text):
        if text is not None:
            self.add('with %s:' % text)
            self._indent += 4
            self._lines_in_context.append(0)
        yield
        if text is not None:
            if self._lines_in_context[-1] == 0:
                self.add('pass')
            self._indent -= 4
            del self._lines_in_context[-1]

    def add(self, text):
        self._lines_in_context[-1] += 1
        self.lines.append((' ' * self._indent) + text)

    def __str__(self):
        return '\n'.join(self.lines)


class Printer(Visitor, Text):
    def __init__(self, factor_prefixes=False, c2_syntax=True):
        super(Visitor, self).__init__()
        super(Text, self).__init__()
        self.factor_prefixes = factor_prefixes
        self.c2_syntax = c2_syntax
        self.c2_net_name = None


def _sanitize_str(s):
    if isinstance(s, text_type):
        sanitized = s
    elif isinstance(s, binary_type):
        sanitized = s.decode('ascii', errors='ignore')
    else:
        sanitized = str(s)
    if len(sanitized) < 64:
        return "'%s'" % sanitized
    else:
        return "'%s'" % sanitized[:64] + '...<+len=%d>' % (len(sanitized) - 64)


def _arg_val(arg):
    if arg.HasField('f'):
        return str(arg.f)
    if arg.HasField('i'):
        return str(arg.i)
    if arg.HasField('s'):
        return _sanitize_str(arg.s)
    if arg.floats:
        return str(list(arg.floats))
    if arg.ints:
        return str(list(arg.ints))
    if arg.strings:
        return str([_sanitize_str(s) for s in arg.strings])
    return '[]'


def commonprefix(m):
    "Given a list of strings, returns the longest common prefix"
    if not m:
        return ''
    s1 = min(m)
    s2 = max(m)
    for i, c in enumerate(s1):
        if c != s2[i]:
            return s1[:i]
    return s1


def format_value(val):
    if isinstance(val, list):
        return '[%s]' % ', '.join("'%s'" % str(v) for v in val)
    else:
        return str(val)


def factor_prefix(vals, do_it):
    vals = [format_value(v) for v in vals]
    prefix = commonprefix(vals) if len(vals) > 1 and do_it else ''
    joined = ', '.join(v[len(prefix):] for v in vals)
    return '%s[%s]' % (prefix, joined) if prefix else joined


def call(op, inputs=None, outputs=None, factor_prefixes=False):
    if not inputs:
        inputs = ''
    else:
        inputs_v = [a for a in inputs if not isinstance(a, tuple)]
        inputs_kv = [a for a in inputs if isinstance(a, tuple)]
        inputs = ', '.join(
            x
            for x in chain(
                [factor_prefix(inputs_v, factor_prefixes)],
                ('%s=%s' % kv for kv in inputs_kv),
            )
            if x
        )
    call = '%s(%s)' % (op, inputs)
    return call if not outputs else '%s = %s' % (
        factor_prefix(outputs, factor_prefixes), call)


def format_device_option(dev_opt):
    if not dev_opt or not (
            dev_opt.device_type or dev_opt.device_id or dev_opt.node_name):
        return None
    return call(
        'DeviceOption',
        [dev_opt.device_type, dev_opt.device_id, "'%s'" % dev_opt.node_name])


@Printer.register(OperatorDef)
def print_op(text, op):
    args = [(a.name, _arg_val(a)) for a in op.arg]
    dev_opt_txt = format_device_option(op.device_option)
    if dev_opt_txt:
        args.append(('device_option', dev_opt_txt))

    if text.c2_net_name:
        text.add(call(
            text.c2_net_name + '.' + op.type,
            [list(op.input), list(op.output)] + args))
    else:
        text.add(call(
            op.type,
            list(op.input) + args,
            op.output,
            factor_prefixes=text.factor_prefixes))
    for arg in op.arg:
        if arg.HasField('n'):
            with text.context('arg: %s' % arg.name):
                text(arg.n)


@Printer.register(NetDef)
def print_net_def(text, net_def):
    if text.c2_syntax:
        text.add(call('core.Net', ["'%s'" % net_def.name], [net_def.name]))
        text.c2_net_name = net_def.name
    else:
        text.add('# net: %s' % net_def.name)
    for op in net_def.op:
        text(op)
    if text.c2_syntax:
        text.c2_net_name = None


@Printer.register(Net)
def print_net(text, net):
    text(net.Proto())


def _get_step_context(step):
    proto = step.Proto()
    if proto.should_stop_blob:
        return call('loop'), False
    if proto.num_iter and proto.num_iter != 1:
        return call('loop', [proto.num_iter]), False
    if proto.num_concurrent_instances > 1:
        return (
            call('parallel',
                 [('num_instances', proto.num_concurrent_instances)]),
            len(step.Substeps()) > 1)
    concurrent = proto.concurrent_substeps and len(step.Substeps()) > 1
    if concurrent:
        return call('parallel'), True
    if proto.report_net:
        return call('run_once'), False
    return None, False


@Printer.register(ExecutionStep)
def print_step(text, step):
    proto = step.Proto()
    step_ctx, do_substep = _get_step_context(step)
    with text.context(step_ctx):
        if proto.report_net:
            with text.context(call('report_net', [proto.report_interval])):
                text(step.get_net(proto.report_net))
        substeps = step.Substeps() + [step.get_net(n) for n in proto.network]
        for substep in substeps:
            sub_proto = (
                substep.Proto() if isinstance(substep, ExecutionStep) else None)
            if sub_proto is not None and sub_proto.run_every_ms:
                substep_ctx = call(
                    'reporter',
                    [str(substep), ('interval_ms', sub_proto.run_every_ms)])
            elif do_substep:
                title = (
                    'workspace'
                    if sub_proto is not None and sub_proto.create_workspace else
                    'step')
                substep_ctx = call(title, [str(substep)])
            else:
                substep_ctx = None
            with text.context(substep_ctx):
                text(substep)
                if proto.should_stop_blob:
                    text.add(call('yield stop_if', [proto.should_stop_blob]))


def _print_task_output(x):
    assert isinstance(x, TaskOutput)
    return 'Output[' + ', '.join(str(x) for x in x.names) + ']'


@Printer.register(Task)
def print_task(text, task):
    outs = ', '.join(_print_task_output(o) for o in task.outputs())
    context = [('node', task.node), ('name', task.name), ('outputs', outs)]
    with text.context(call('Task', context)):
        text(task.get_step())


@Printer.register(TaskGroup)
def print_task_group(text, tg, header=None):
    with text.context(header or call('TaskGroup')):
        for task in tg.tasks_by_node().tasks():
            text(task)


@Printer.register(Job)
def print_job(text, job):
    text(job.init_group, 'Job.current().init_group')
    text(job.epoch_group, 'Job.current().epoch_group')
    with text.context('Job.current().stop_conditions'):
        for out in job.stop_conditions:
            text.add(_print_task_output(out))
    text(job.download_group, 'Job.current().download_group')
    text(job.exit_group, 'Job.current().exit_group')


def to_string(obj, **kwargs):
    """
    Given a Net, ExecutionStep, Task, TaskGroup or Job, produces a string
    with detailed description of the execution steps.
    """
    printer = Printer(**kwargs)
    printer(obj)
    return str(printer)


def debug_net(net):
    """
    Given a Net, produce another net that logs info about the operator call
    before each operator execution. Use for debugging purposes.
    """
    assert isinstance(net, Net)
    debug_net = Net(str(net))
    assert isinstance(net, Net)
    for op in net.Proto().op:
        text = Text()
        print_op(op, text)
        debug_net.LogInfo(str(text))
        debug_net.Proto().op.extend([op])
    return debug_net
